from abc import ABC, abstractmethod
from collections import deque
from functools import cached_property

import typing_extensions as typing

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.node.base import IsNode, get_nodes
from telegrinder.tools.fullname import fullname
from telegrinder.types.objects import Update

type CheckResult = bool | typing.Awaitable[bool]


class ABCRule(ABC):
    requires: deque["ABCRule"] = deque()

    if typing.TYPE_CHECKING:

        @abstractmethod
        def check(self, *args: typing.Any, **kwargs: typing.Any) -> CheckResult:
            pass
    else:

        @abstractmethod
        def check(self, *args, **kwargs):
            pass

    def __init_subclass__(
        cls,
        *,
        requires: typing.Iterable["ABCRule"] | None = None,
    ) -> None:
        """Merges requirements from inherited classes and rule-specific requirements."""
        requirements = list[ABCRule]()
        for base in cls.__mro__:
            if issubclass(base, ABCRule) and base != cls:
                requirements.extend(base.requires or ())

        requirements.extend(requires or ())
        cls.requires = deque(dict.fromkeys(requirements))

    def __and__(self, other: object, /) -> "AndRule":
        if not isinstance(other, ABCRule):
            return NotImplemented
        return AndRule(self, other)

    def __iadd__(self, other: object, /) -> "AndRule":
        return self.__and__(other)

    def __or__(self, other: object, /) -> "OrRule":
        if not isinstance(other, ABCRule):
            return NotImplemented
        return OrRule(self, other)

    def __ior__(self, other: object, /) -> "OrRule":
        return self.__or__(other)

    def __invert__(self) -> "NotRule":
        return NotRule(self)

    def __repr__(self) -> str:
        return "<{}, requires={!r}>".format(fullname(self), self.requires)

    @cached_property
    def required_nodes(self) -> dict[str, IsNode]:
        return get_nodes(self.check)

    def as_optional(self) -> "ABCRule":
        return self | Always()

    def should_fail(self) -> "ABCRule":
        return self & Never()


class AndRule(ABCRule):
    def __init__(self, *rules: ABCRule) -> None:
        self.rules = rules

    async def check(self, event: Update, api: API, ctx: Context) -> bool:
        ctx_copy = ctx.copy()

        for rule in self.rules:
            if not await check_rule(api, rule, event, ctx_copy):
                return False

        ctx |= ctx_copy
        return True


class OrRule(ABCRule):
    def __init__(self, *rules: ABCRule) -> None:
        self.rules = rules

    async def check(self, event: Update, api: API, ctx: Context) -> bool:
        for rule in self.rules:
            ctx_copy = ctx.copy()

            if await check_rule(api, rule, event, ctx_copy):
                ctx |= ctx_copy
                return True

        return False


class NotRule(ABCRule):
    def __init__(self, rule: ABCRule) -> None:
        self.rule = rule

    async def check(self, event: Update, api: API, ctx: Context) -> bool:
        ctx_copy = ctx.copy()
        return not await check_rule(api, self.rule, event, ctx_copy)


class Never(ABCRule):
    """Neutral element for `|` (OrRule)."""

    def check(self) -> typing.Literal[False]:
        return False


class Always(ABCRule):
    """Neutral element for `&` (AndRule)."""

    def check(self) -> typing.Literal[True]:
        return True


__all__ = (
    "ABCRule",
    "Always",
    "AndRule",
    "CheckResult",
    "Never",
    "NotRule",
    "OrRule",
    "check_rule",
)



import typing

import aiosqlite  # type: ignore

from telegrinder.modules import logger
from telegrinder.node import NodeScope, scalar_node


@scalar_node(scope=NodeScope.PER_CALL)
class DB:
    @classmethod
    async def compose(cls) -> typing.AsyncGenerator[aiosqlite.Connection, None]:  # type: ignore
        connection = await aiosqlite.connect("test.db")  # type: ignore
        logger.info("Opening connection")
        yield connection
        logger.info("Closing connection")
        await connection.close()  # type: ignore


async def create_tables() -> None:
    async with aiosqlite.connect("test.db") as conn:  # type: ignore
        await conn.execute(  # type: ignore
            "create table if not exists admins(id integer primary key autoincrement, telegram_id text unique)"
        )



from telegrinder.bot.dispatch.return_manager import InlineQueryReturnManager
from telegrinder.bot.dispatch.view.base import BaseView
from telegrinder.types.enums import UpdateType


class InlineQueryView(BaseView):
    def __init__(self) -> None:
        super().__init__(UpdateType.INLINE_QUERY)
        self.return_manager = InlineQueryReturnManager()


__all__ = ("InlineQueryView",)



from telegrinder.bot.dispatch.return_manager.callback_query import CallbackQueryReturnManager
from telegrinder.bot.dispatch.view.base import BaseView
from telegrinder.types.enums import UpdateType


class CallbackQueryView(BaseView):
    def __init__(self) -> None:
        super().__init__(UpdateType.CALLBACK_QUERY)
        self.return_manager = CallbackQueryReturnManager()


__all__ = ("CallbackQueryView",)



import re

PATTERN = re.compile(r"[^ a-z A-Z 0-9 \s]")


def to_pascal_case(string: str, /) -> str:
    return "".join(
        "".join(char.capitalize() for char in sub_str)
        for sub_str in (char.split() for char in PATTERN.split(string))
    )


__all__ = ("to_pascal_case",)



import typing
from datetime import timedelta
from functools import wraps

from telegrinder.tools.formatting.deep_links.parsing import (
    NO_VALUE,
    DeepLinkFunction,
    NoValue,
    Parameter,
    get_query_params,
    parse_deep_link,
)
from telegrinder.tools.formatting.deep_links.validators import separate_by_plus_char

type Permission = typing.Literal[
    "change_info",
    "post_messages",
    "edit_messages",
    "delete_messages",
    "invite_users",
    "restrict_members",
    "promote_members",
    "pin_messages",
    "manage_topics",
    "manage_video_chats",
    "anonymous",
    "manage_chat",
    "post_stories",
    "edit_stories",
    "delete_stories",
]
type Peer = typing.Literal[
    "users",
    "bots",
    "groups",
    "channels",
]


def deep_link[**P](
    link: str,
    /,
    *,
    no_value_params: set[str] | None = None,
    order_params: set[str] | None = None,
) -> typing.Callable[[DeepLinkFunction[P]], DeepLinkFunction[P]]:
    if not link.startswith("tg://"):
        raise RuntimeError("Invalid deep link format. The link must start with 'tg://'.")

    def inner(func: DeepLinkFunction[P]) -> DeepLinkFunction[P]:
        @wraps(func)
        def wrapper(*_: P.args, **kwargs: P.kwargs) -> str:
            return parse_deep_link(
                link=link,
                params=get_query_params(func, kwargs, order_params),
                no_value_params=no_value_params,
            )

        return wrapper

    return inner


@deep_link("tg://resolve")
def tg_public_username_link(
    *,
    username: Parameter[str, "domain"],
    draft_text: Parameter[str | None, "text"] = None,
    open_profile: Parameter[bool, "profile"] = False,
) -> str:
    """Used to link to public `users`, `groups` and `channels`.

    :param username: Username.
    :param draft_text: Optional. UTF-8 text to pre-enter into the text input bar, if the user can write in the chat.
    :param open_profile: Optional. If set, clicking on this link should open the destination peer's profile page, not the chat view.
    """
    ...


@deep_link("tg://user")
def tg_mention_link(*, user_id: Parameter[int, "id"]) -> str:
    """ID links are merely an abstraction offered by the `Bot API` to simplify construction of
    `inputMessageEntityMentionName` and `inputKeyboardButtonUserProfile` constructors, and should be
    ignored by normal clients.

    :param user_id: User ID.
    """
    ...


@deep_link("tg://openmessage")
def tg_open_message_link(
    *,
    chat_id: int | None = None,
    user_id: int | None = None,
    message_id: int | None = None,
) -> str: ...


@deep_link("tg://emoji")
def tg_emoji_link(*, emoji_id: Parameter[int, "id"]) -> str:
    """Emoji links are merely an abstraction offered by the `Bot API` to simplify construction of
    `messageEntityCustomEmoji` constructors, and should be ignored by normal clients.

    :param emoji_id: Custom emoji ID.
    """
    ...


@deep_link("tg://addemoji")
def tg_emoji_stickerset_link(*, short_name: Parameter[str, "set"]) -> str:
    """Used to import custom emoji stickersets.

    :param short_name: Stickerset short name, used when installing stickers.
    """
    ...


@deep_link("tg://resolve")
def tg_story_link(
    *,
    username: Parameter[str, "domain"],
    story_id: Parameter[int, "story"],
) -> str:
    """Used to link to a Telegram Story.

    :param username: Username of the user or channel that posted the story.
    :param story_id: ID of the Telegram Story.
    """
    ...


@deep_link("tg://join")
def tg_chat_invite_link(*, invite_hash: Parameter[str, "invite"]) -> str:
    """Used to invite users to private `groups` and `channels`.

    :param invite_hash: Invite hash.
    """
    ...


@deep_link("tg://addlist")
def tg_chat_folder_link(*, slug: str) -> str:
    """Used to add chat folders.

    :param slug: Folder slug.
    """
    ...


@deep_link("tg://resolve")
def tg_public_message_link(
    *,
    dialog: Parameter[str, "domain"],
    message_id: Parameter[int, "post"],
    single: bool = False,
    thread_id: Parameter[int | None, "thread"] = None,
    comment: str | None = None,
    timestamp: Parameter[timedelta | None, "t"] = None,
) -> str:
    """Used to link to specific messages in public or private `groups` and `channels`.

    :param dialog: Dialog username.
    :param message_id: Message ID.
    :param single: Optional. For albums/grouped media, if set indicates that this is a link to a specific media in the album; otherwise, it is a link to the entire album.
    :param thread_id: Optional. For message threads, contains the thread ID.
    :param comment: Optional. For channel comments, username will contain the channel username, id will contain the message ID of the channel message that started the comment section and this field will contain the message ID of the comment in the discussion group.
    :param timestamp: Optional. Timestamp at which to start playing the media file present in the body or in the webpage preview of the message.
    """
    ...


@deep_link("tg://privatepost")
def tg_private_message_link(
    *,
    channel: str,
    message_id: Parameter[int, "post"],
    single: bool = False,
    thread_id: Parameter[int | None, "thread"] = None,
    comment: str | None = None,
    timestamp: Parameter[timedelta | None, "t"] = None,
) -> str:
    """Used to link to specific messages in public or private `groups` and `channels`.

    :param channel: Channel or supergroup ID.
    :param post: Message ID.
    :param single: Optional. For albums/grouped media, if set indicates that this is a link to a specific media in the album; otherwise, it is a link to the entire album.
    :param thread: Optional. For message threads, contains the thread ID.
    :param comment: Optional. For channel comments, username will contain the channel username, id will contain the message ID of the channel message that started the comment section and this field will contain the message ID of the comment in the discussion group.
    :param t: Optional. Timestamp at which to start playing the media file present in the body or in the webpage preview of the message.
    """
    ...


@deep_link("tg://msg_url")
def tg_share_link(*, url: str, text: str | None = None) -> str:
    """Used to share a prepared message and URL into a chosen chat's text field.

    These links should be handled as follows:
    * Open a dialog selection prompt
    * After selection: validate, trim and enter the URL at the beginning of the text field
    * Append a newline to the text field
    * Append and select the `text`, if present

    :param url: URL to share (`urlencoded`).
    :param text: Optional. Text to prepend to the URL.
    """
    ...


@deep_link("tg://boost")
def tg_public_channel_boost_link(*, channel_username: Parameter[str, "domain"]) -> str:
    """Used by users to boost public channels, granting them the ability to post stories and further perks.

    :param channel_username: Channel username.
    """
    ...


@deep_link("tg://boost")
def tg_private_channel_boost_link(*, channel_id: Parameter[int, "channel"]) -> str:
    """Used by users to boost private channels, granting them the ability to post stories and further perks.

    :param channel_id: Channel ID.
    """
    ...


@deep_link("tg://invoice")
def tg_invoice_link(*, slug: str) -> str:
    """Used to initiate payment of an invoice, generated using `payments.exportedInvoice`.

    :param slug: The invoice slug to be used during payment.
    """
    ...


@deep_link("tg://setlanguage")
def tg_language_pack_link(
    *,
    lang_pack: Parameter[str | None, "lang"] = None,
) -> str:
    """Used to import custom language packs using `langpack.getLangPack`.

    :param lang_pack: Optional. Name of language pack.
    """
    ...


@deep_link("tg://premium_multigift")
def tg_premium_multigift_link(*, ref: str) -> str:
    """Used to bring the user to the screen used for gifting `Telegram Premium` subscriptions to friends,
    see here for more info on gifting `Telegram Premium` to multiple users.

    This link is used to invite users to gift `Premium` subscription to other users, see here for the different
    link type containing the actual giftcodes that can be used to import a gifted `Telegram Premium` subscription.

    :param ref: Used by official apps for analytics using `help.saveAppLog`.
    """
    ...


@deep_link("tg://premium_offer")
def tg_premium_offer_link(*, ref: str | None = None) -> str:
    """Used by official apps to show the `Telegram Premium` subscription page.

    :param ref: Optional. Used by official apps for analytics using `help.saveAppLog`.
    """
    ...


@deep_link("tg://resolve")
def tg_bot_start_link(
    *,
    bot_username: Parameter[str, "domain"],
    parameter: Parameter[str | None, "start"] = None,
) -> str:
    """Used to link to bots.

    :param bot_username: Bot username.
    :param parameter: Optional. Start parameter, up to 64 `base64url` characters: if provided and the `bot_username` is indeed a bot, \
    the text input bar should be replaced with a `Start` button (even if the user has already started the bot) that should invoke \
    `messages.startBot` with the appropriate `parameter` once clicked. Note that if the `bot_username` is equal to the `premium_bot_username` \
    configuration value, clicking on this link should immediately invoke `messages.startBot` with the appropriate `parameter`.
    """
    ...


@deep_link("tg://resolve")
def tg_bot_startgroup_link(
    *,
    bot_username: Parameter[str, "domain"],
    parameter: Parameter[str | NoValue, "startgroup"] = NO_VALUE,
    permissions: Parameter[list[Permission] | None, "admin", separate_by_plus_char] = None,
) -> str:
    """Used to add bots to groups.
    First of all, check that the `<bot_username>` indeed links to a bot.

    Then, for group links:

    If the `admin` parameter is not provided:
        - Bring up a dialog selection of groups where the user can add members
        - Add the bot to the group
        - If a `parameter` is provided, invoke `messages.startBot` with the appropriate `parameter`

    If the admin parameter is provided:
        - Bring up a dialog selection of groups where the user can add/edit admins
        - If the bot is already an admin of the group, combine existing admin rights with the admin rights in admin
        - Add the bot as admin/modify admin permissions to the new rights
        - If a parameter is provided, invoke `messages.startBot` with the appropriate `parameter`

    :param bot_username: Bot username.
    :param parameter: Optional. Start parameter, only for group links, up to 64 `base64url` characters: if provided and the `bot_username` is indeed a bot, \
    `messages.startBot` with the appropriate parameter should be invoked after adding the bot to the group.
    :param permissions: Optional. A combination of the following identifiers separated by `+`, each corresponding to the appropriate flag in the `chatAdminRights` constructor.
    """
    ...


@deep_link("tg://resolve", no_value_params={"startchannel"})
def tg_bot_startchannel_link(
    *,
    bot_username: Parameter[str, "domain"],
    permissions: Parameter[list[Permission], "admin", separate_by_plus_char],
) -> str:
    """Used to add bots to channels.
    First of all, check that the `<bot_username>` indeed links to a bot.

    For channel links:
    - Bring up a dialog selection of channels where the user can add/edit admins
    - If the bot is already an admin of the channel, combine existing admin rights with the admin rights in admin
    - Add the bot as admin/modify admin permissions to the new rights

    :param bot_username: Bot username.
    :param permissions: A combination of the following identifiers separated by `+`, each corresponding to the \
    appropriate flag in the `chatAdminRights` constructor.
    """
    ...


@deep_link("tg://resolve")
def tg_main_mini_app_link(
    *,
    bot_username: Parameter[str, "domain"],
    parameter: Parameter[str | NoValue, "startapp"] = NO_VALUE,
    mode: typing.Literal["compact"] | None = None,
) -> str:
    """Used to open `Main Mini Apps`.

    If the specified bot does not have a configured `Main` Mini App (i.e. the user.bot_has_main_app flag will not be set),
    fall back to the behavior of username links.

    The main mini app should be opened using `messages.requestMainWebView`.

    :param bot_username: Bot username.
    :param parameter: Optional. If provided, should be passed to `messages.requestMainWebView.start_param`.
    :param mode: Optional. If equal to `compact`, the `messages.requestMainWebView.compact` flag must be set.
    """
    ...


@deep_link("tg://resolve")
def tg_direct_mini_app_link(
    *,
    bot_username: Parameter[str, "domain"],
    short_name: Parameter[str, "appname"],
    parameter: Parameter[str | NoValue, "startapp"] = NO_VALUE,
    mode: typing.Literal["compact"] | None = None,
) -> str:
    """Used to share `Direct` link Mini apps.

    These links are different from bot attachment menu deep links, because they don't require the user to install
    an attachment menu, and a single bot can offer multiple named mini apps, distinguished by their `short_name`.

    These links should be handled as specified in the direct link Mini Apps documentation.

    :param bot_username: Username of the bot that owns the game.
    :param short_name: Mini app short name, to pass to `inputBotAppShortName.short_name` when invoking `messages.getBotApp`.
    :param parameter: Optional. To pass to `messages.requestAppWebView.start_param`.
    :param mode: Optional. If equal to `compact`, the `messages.requestAppWebView.compact` flag must be set.
    """
    ...


@deep_link("tg://resolve")
def tg_bot_attach_open_current_chat(
    *,
    bot_username: Parameter[str, "domain"],
    parameter: Parameter[str | NoValue, "startattach"] = NO_VALUE,
) -> str:
    """After installing the attachment/side menu entry globally, opens the associated mini app using
    `messages.requestWebView` in the currently open chat, by passing it to the peer parameter of `messages.requestWebView`.

    If the current chat is not supported by the `attachMenuBot.peer_types` field:

    - If the user has just installed the attachment menu in the previous step, notify the user that the attachment menu was installed successfully.
    - Otherwise, notify the user that the attachment menu webapp can't be opened in the specified chat.

    :param bot_username: Username of the bot that owns the attachment/side menu entry.
    :param parameter: Optional. If provided, should be passed to `messages.requestWebView.start_param`.
    """
    ...


@deep_link("tg://resolve")
def tg_bot_attach_open_specific_chat(
    *,
    username: Parameter[str, "domain"],
    phone_number: Parameter[str, "phone"],
    bot_username: Parameter[str, "attach"],
    parameter: Parameter[str | NoValue, "startattach"] = NO_VALUE,
) -> str:
    """After installing the `attachment/side` menu entry globally, opens the associated mini app using
    `messages.requestWebView` in a specific chat (passed to the peer parameter of `messages.requestWebView`).

    If the specified chat is not supported by the attachMenuBot.peer_types field:

    - If the user has just installed the attachment menu in the previous step, notify the user that the attachment menu was installed successfully.
    - Otherwise, notify the user that the attachment menu webapp can't be opened in the specified chat.
    """
    ...


@deep_link("tg://resolve")
def tg_bot_attach_open_any_chat(
    *,
    bot_username: Parameter[str, "domain"],
    parameter: Parameter[str | NoValue, "startattach"] = NO_VALUE,
    peer: Parameter[list[Peer] | None, "choose", separate_by_plus_char] = None,
) -> str:
    """After installing the `attachment/side` menu entry globally, opens a dialog selection form that will open the attachment menu
    mini app using `messages.requestWebView` in a specific chat (pass it to the peer parameter of `messages.requestWebView`).

    :param bot_username: Username of the bot that owns the `attachment/side` menu.
    :param parameter: Optional. If provided, should be passed to `messages.requestWebView.start_param`.
    :param peer: Optional. A combination of `users`, `bots`, `groups`, `channels` separated by `+`: indicates the dialog types to show in \
    the dialog selection popup: must be intersected with the dialog types contained in the `attachMenuBot.peer_types` field before use.
    """
    ...


__all__ = (
    "tg_bot_attach_open_any_chat",
    "tg_bot_attach_open_current_chat",
    "tg_bot_attach_open_specific_chat",
    "tg_bot_start_link",
    "tg_bot_startchannel_link",
    "tg_bot_startgroup_link",
    "tg_chat_folder_link",
    "tg_chat_invite_link",
    "tg_direct_mini_app_link",
    "tg_emoji_link",
    "tg_emoji_stickerset_link",
    "tg_invoice_link",
    "tg_language_pack_link",
    "tg_main_mini_app_link",
    "tg_mention_link",
    "tg_open_message_link",
    "tg_premium_multigift_link",
    "tg_premium_offer_link",
    "tg_private_channel_boost_link",
    "tg_private_message_link",
    "tg_public_channel_boost_link",
    "tg_public_message_link",
    "tg_public_username_link",
    "tg_share_link",
    "tg_story_link",
)



import dataclasses
import typing

from telegrinder.tools.fullname import fullname


class BaseButton:
    def get_data(self) -> dict[str, typing.Any]:
        assert dataclasses.is_dataclass(self), f"{fullname(self)} is not a dataclass."
        return {k: v for k, v in dataclasses.asdict(self).items() if v is not None}


@dataclasses.dataclass(kw_only=True)
class BaseStaticButton(BaseButton):
    row: bool = dataclasses.field(default=False, repr=False)


__all__ = ("BaseButton", "BaseStaticButton")



import asyncio
import dataclasses
import datetime

from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Markup, Text

api = API(token=Token.from_env())
bot = Telegrinder(api)
RAVIOLI_TIME_TO_COOK = 9 * 60

logger.set_level("DEBUG")


@dataclasses.dataclass
class TimerInfo:
    name: str
    end_time: datetime.datetime


class DummyDB:
    def __init__(self) -> None:
        self.storage: dict[int, list[TimerInfo]] = {}

    def get(self, user_id: int) -> list[TimerInfo]:
        return self.storage.get(user_id) or []

    def add(self, user_id: int, tinfo: TimerInfo) -> None:
        if user_id not in self.storage:
            self.storage[user_id] = []
        self.storage[user_id].append(tinfo)

    async def new(self, name: str, time: int, message: Message) -> TimerInfo:
        await bot.loop_wrapper.create_task(self.timer(name, time, message))
        return TimerInfo(name, datetime.datetime.now() + datetime.timedelta(seconds=time))

    async def timer(self, name: str, time: int, message: Message) -> None:
        await asyncio.sleep(time)
        await message.reply(f"{name}'s are ready! Switch off the oven quickly")
        self.storage[message.from_user.id].pop(0)


db = DummyDB()


@bot.on.message(Markup(["/ravioli", "/ravioli <ravioli_name>"]))
async def start(message: Message, ravioli_name: str = "Ravioli") -> str:
    ravioli = await db.new(ravioli_name, RAVIOLI_TIME_TO_COOK, message)
    db.add(message.from_user.id, ravioli)
    return f"Timer for ravioli {ravioli_name!r} is set!"


@bot.on.message(Text("/cooking"))
async def cooking(message: Message):
    boiling = db.get(message.from_user.id)
    if not boiling:
        await message.answer("Nothing is cooking right now!")
        return
    text = "\n".join(
        "{}. {} will be ready at {}".format(
            index,
            record.name,
            record.end_time.strftime("%m.%d.%Y %H:%M"),
        )
        for index, record in enumerate(boiling, start=1)
    )
    await message.answer("Boiling raviolies:\n\n" + text)


bot.run_forever(skip_updates=True)



from __future__ import annotations

import asyncio
import pathlib
from functools import cached_property, wraps
from http import HTTPStatus

import msgspec
import typing_extensions as typing
from fntypes.misc import is_ok
from fntypes.result import Error, Ok, Result

from telegrinder.api.error import APIError
from telegrinder.api.response import APIResponse
from telegrinder.api.token import Token
from telegrinder.client import ABCClient, AiohttpClient, MultipartFormProto
from telegrinder.client.aiohttp import DEFAULT_TIMEOUT
from telegrinder.model import decoder
from telegrinder.types.methods import APIMethods

type Json = str | int | float | bool | list[Json] | dict[str, Json] | None
type Data = dict[str, typing.Any]
type Files = dict[str, tuple[str, bytes]]
type APIRequestMethod[T: API, **P, R] = typing.Callable[
    typing.Concatenate[T, P],
    typing.Coroutine[typing.Any, typing.Any, Result[R, APIError]],
]

DEFAULT_MAX_RETRIES: typing.Final[int] = 5


def compose_data(
    client: ABCClient,
    data: Data,
    files: Files,
) -> MultipartFormProto:
    if not data and not files:
        return client.multipart_form_factory()
    return client.get_form(data=data, files=files)


def retryer[T: API, **P, R](func: APIRequestMethod[T, P, R], /) -> APIRequestMethod[T, P, R]:
    @wraps(func)
    async def wrapper(
        self: T,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Result[typing.Any, APIError]:
        retries_counter = 0

        while True:
            result = await func(self, *args, **kwargs)

            if is_ok(result) or not self.retryer_enabled or retries_counter >= self.max_retries:
                return result

            if result.error.status_code == HTTPStatus.TOO_MANY_REQUESTS:
                await asyncio.sleep(result.error.retry_after.unwrap_or(5.0))

            elif result.error.status_code == HTTPStatus.INTERNAL_SERVER_ERROR and "restart" in result.error.error:
                await asyncio.sleep(10.0)

            elif result.error.migrate_to_chat_id:
                kwargs["chat_id"] = result.error.migrate_to_chat_id.value

            retries_counter += 1

    return wrapper  # type: ignore


class API(APIMethods):
    """Bot API with available API methods and http client."""

    http: ABCClient

    API_URL = "https://api.telegram.org/"
    API_FILE_URL = "https://api.telegram.org/file/"

    def __init__(
        self,
        token: Token,
        *,
        http: ABCClient | None = None,
        enable_retryer: bool = True,
        max_retries: int = DEFAULT_MAX_RETRIES,
    ) -> None:
        self.token = token
        self.http = http or AiohttpClient()
        self._retryer_enabled = enable_retryer
        self._max_retries = max_retries
        super().__init__(api=self)

    def __repr__(self) -> str:
        return "<{}: id={}, http={!r}, retryer_enabled={}, max_retries={}>".format(
            type(self).__name__,
            self.id,
            self.http,
            self._retryer_enabled,
            self._max_retries,
        )

    @cached_property
    def id(self) -> int:
        return self.token.bot_id

    @property
    def request_url(self) -> str:
        return self.API_URL + f"bot{self.token}/"

    @property
    def request_file_url(self) -> str:
        return self.API_FILE_URL + f"bot{self.token}/"

    @property
    def retryer_enabled(self) -> bool:
        return self._retryer_enabled

    @property
    def max_retries(self) -> int:
        return self._max_retries

    async def download_file(
        self,
        file_path: str | pathlib.Path,
        timeout: int | float = DEFAULT_TIMEOUT,
    ) -> bytes:
        return await self.http.request_content(
            url=f"{self.request_file_url}/{file_path}",
            timeout=timeout,
        )

    @retryer
    async def request(
        self,
        method: str,
        data: Data | None = None,
        files: Files | None = None,
        **kwargs: typing.Any,
    ) -> Result[Json, APIError]:
        """Request a `JSON` response using http method `POST` and passing data, files as `multipart/form-data`."""
        response = await self.http.request_json(
            url=self.request_url + method,
            method="POST",
            data=compose_data(self.http, data or {}, files or {}),
            **kwargs,
        )
        if response.get("ok", False) is True:
            return Ok(response["result"])
        return Error(
            APIError(
                code=response.get("error_code", 400),
                error=response.get("description", "Something went wrong"),
                data=response.get("parameters", {}),
            ),
        )

    @retryer
    async def request_raw(
        self,
        method: str,
        data: Data | None = None,
        files: Files | None = None,
        **kwargs: typing.Any,
    ) -> Result[msgspec.Raw, APIError]:
        """Request a `raw` response using http method `POST` and passing data, files as `multipart/form-data`."""
        response_bytes = await self.http.request_bytes(
            url=self.request_url + method,
            method="POST",
            data=compose_data(self.http, data or {}, files or {}),
            **kwargs,
        )
        return decoder.decode(response_bytes, type=APIResponse).to_result()


__all__ = ("API", "compose_data", "retryer")



import pytest

from telegrinder.types.objects import Update

UPDATE = Update.from_raw(
    b"""
    {
        "update_id": 12345,
        "message": {
            "message_id": 23423522,
            "from": {
                "id": 123,
                "is_bot": false,
                "first_name": "John",
                "last_name": "Doe",
                "username": "Johndoe333",
                "language_code": "en"
            },
            "chat": {
                "id": 123,
                "first_name": "John",
                "last_name": "Doe",
                "is_bot": false,
                "username": "Johndoe333",
                "type": "private"
            },
            "date": 1234567898,
            "text": "Hello, Telegrinder! btw, laurelang - nice pure logical programming launguage ^_^"
        }
    }
    """
)


@pytest.fixture()
def message_update():
    return UPDATE



import typing

from telegrinder.bot.cute_types.pre_checkout_query import PreCheckoutQueryCute
from telegrinder.bot.dispatch.return_manager.abc import BaseReturnManager, register_manager


class PreCheckoutQueryManager(BaseReturnManager):
    @register_manager(bool)
    @staticmethod
    async def bool_manager(value: bool, event: PreCheckoutQueryCute) -> None:
        await event.answer(value)

    @register_manager(dict)
    @staticmethod
    async def dict_manager(value: dict[str, typing.Any], event: PreCheckoutQueryCute) -> None:
        await event.answer(**value)


__all__ = ("PreCheckoutQueryManager",)



import random

from telegrinder import (
    MESSAGE_IN_CHAT,
    CallbackQuery,
    Checkbox,
    Dispatch,
    Message,
    MessageReplyHandler,
    WaiterMachine,
)
from telegrinder.bot.dispatch.waiter_machine.hasher.callback import CALLBACK_QUERY_FOR_MESSAGE
from telegrinder.rules import (
    FuzzyText,
    HasText,
    IntegerInRange,
    PayloadEqRule,
    PayloadMarkupRule,
    Text,
)
from telegrinder.tools.formatting import HTMLFormatter, block_quote, link
from telegrinder.tools.keyboard import InlineButton, InlineKeyboard

dp = Dispatch()
wm = WaiterMachine(dp)


kb = (
    InlineKeyboard()
    .add(InlineButton("✅ Confirm", callback_data="action/confirm"))
    .row()
    .add(InlineButton("🛰 Webhooks", callback_data="action/webhooks"))
    .add(InlineButton("✍️ Quote", callback_data="action/quote"))
    .add(InlineButton("🎲 Guess the number", callback_data="action/guess"))
    .row()
    .add(InlineButton("❌ Cancel", callback_data="action/cancel"))
).get_markup()


@dp.message(Text("/start") | FuzzyText("hello"))
async def start(message: Message) -> None:
    me = (await message.ctx_api.get_me()).unwrap().full_name
    await message.answer(f"Hello, {message.from_user.full_name}, im {me} and i work on a webhook server!")


@dp.message(Text("/cars"))
async def car_choice(message: Message) -> None:
    picked, m_id = await (
        Checkbox(wm, message.chat.id, "🚘 Choose no more than three cars", max_in_row=2)
        .add_option("bentley", "Bentley Continental", "Continental 🤍")
        .add_option("mazda", "Mazda rx 7", "Mazda rx 7 🩵")
        .add_option("toyota", "Toyota Supra mk5", "Supra mk5 💜")
        .wait(CALLBACK_QUERY_FOR_MESSAGE, message.ctx_api)
    )
    await message.edit(
        "🚘 You picked: {}.".format(", ".join(c for c in picked if picked[c])),
        message_id=m_id,
    )


@dp.message(Text("/menu"))
async def handle_menu_command(message: Message) -> None:
    await message.answer("📋 Menu:", reply_markup=kb)


@dp.callback_query(PayloadEqRule("action/webhooks"))
async def handle_query_webhook(cb: CallbackQuery) -> None:
    await cb.answer()
    await cb.ctx_api.send_message(
        text=HTMLFormatter(
            link(
                "https://core.telegram.org/bots/webhooks",
                text="🛰 Marvin's Marvellous Guide to All Things Webhook.",
            )
        ),
        chat_id=cb.chat_id.unwrap(),
        parse_mode=HTMLFormatter.PARSE_MODE,
    )


@dp.callback_query(PayloadEqRule("action/quote"))
async def handle_query_quote(cb: CallbackQuery) -> None:
    await cb.answer()
    message = (
        await cb.ctx_api.send_message(
            text="✍️ Send me any message and i'll quote it!",
            chat_id=cb.chat_id.unwrap(),
        )
    ).unwrap()
    msg, _ = await wm.wait(
        *MESSAGE_IN_CHAT(message.chat.id),
        release=HasText(),
        on_miss=MessageReplyHandler("Im still waiting for your message!"),
    )
    await msg.reply(HTMLFormatter(block_quote(msg.text.unwrap())), parse_mode=HTMLFormatter.PARSE_MODE)


@dp.callback_query(PayloadEqRule("action/guess"))
async def handle_query_guess(cb: CallbackQuery) -> None:
    await cb.answer()
    message = (
        await cb.ctx_api.send_message(
            text="🎲 Okay, i guessed a number between 1 and 10!",
            chat_id=cb.chat_id.unwrap(),
        )
    ).unwrap()
    msg, _ = await wm.wait(
        MESSAGE_IN_CHAT,
        message.chat.id,
        release=IntegerInRange(range(1, 11)),
        on_miss=MessageReplyHandler("Send a number between 1 and 10!"),
    )
    random_number = random.randint(1, 10)
    if int(msg.text.unwrap()) == random_number:
        await msg.answer("🎲 Yayyyyy you guessed it!")
        return
    await msg.answer(f"🎲 Ohh noooo...  i guessed the number {random_number} :-(")


@dp.callback_query(PayloadMarkupRule("action/<action>"))
async def handle_query_action(cb: CallbackQuery, action: str) -> None:
    await cb.answer("✅" if action == "confirm" else "❌")
    match action:
        case "confirm":
            await cb.edit_text("✅ Confirmed!")
        case "cancel":
            await cb.delete()



from __future__ import annotations

import abc
import dataclasses
import enum

from fntypes.option import Option

from telegrinder.bot.rules.state import State, StateMeta


@dataclasses.dataclass
class StateData[Payload]:
    key: str | enum.Enum
    payload: Payload

    async def save(self, user_id: int, storage: ABCStateStorage[Payload]) -> None:
        await storage.set(user_id, key=self.key, payload=self.payload)


class ABCStateStorage[Payload](abc.ABC):
    @abc.abstractmethod
    async def get(self, user_id: int) -> Option[StateData[Payload]]: ...

    @abc.abstractmethod
    async def delete(self, user_id: int) -> None: ...

    @abc.abstractmethod
    async def set(self, user_id: int, key: str | enum.Enum, payload: Payload) -> None: ...

    def State(self, key: str | StateMeta | enum.Enum = StateMeta.ANY, /) -> State[Payload]:  # noqa: N802
        """Can be used as a shortcut to get a state rule dependant on current storage."""
        return State(storage=self, key=key)


__all__ = ("ABCStateStorage", "StateData")



from telegrinder import (
    API,
    ABCRule,
    Message,
    Telegrinder,
    Token,
)
from telegrinder.bot import MESSAGE_FROM_USER, WaiterMachine
from telegrinder.modules import logger
from telegrinder.rules import Text

api = API(token=Token.from_env())
bot = Telegrinder(api)
wm = WaiterMachine(bot.dispatch)

logger.set_level("INFO")


class HasPhoto(ABCRule):
    async def check(self, message: Message) -> bool:
        return bool(message.photo and message.photo.unwrap())


class HasNicePhoto(ABCRule, requires=[HasPhoto()]):
    async def check(self, message: Message) -> bool:
        return message.photo.unwrap()[0].width > message.photo.unwrap()[0].height


@bot.on.message(Text("/chain"))
async def start_handler(m: Message):
    await m.answer("Send me a photo please")
    m, _ = await wm.wait(
        hasher=MESSAGE_FROM_USER(m.from_user.id),
        release=HasPhoto(),
    )
    await m.reply("Great photo! Chain completed.")


@bot.on.message(HasNicePhoto())
async def dad_photo_handler(m: Message):
    await m.reply("Wow this photo is really nice")


@bot.on.message(HasPhoto())
async def photo_handler(m: Message):
    await m.answer("You have a photo!")


bot.run_forever()



import pathlib
import random

from telegrinder import (
    API,
    CALLBACK_QUERY_FOR_MESSAGE,
    MESSAGE_FROM_USER_IN_CHAT,
    Context,
    InlineButton,
    InlineKeyboard,
    Message,
    Telegrinder,
    Token,
)
from telegrinder.bot import WaiterMachine
from telegrinder.bot.dispatch.handler import MessageReplyHandler
from telegrinder.bot.dispatch.middleware import ABCMiddleware
from telegrinder.bot.rules.is_from import IsPrivate, IsUser
from telegrinder.modules import logger
from telegrinder.node import Me, UserId, as_node
from telegrinder.rules import (
    CallbackDataEq,
    FuzzyText,
    HasText,
    IsUpdateType,
    Markup,
    Text,
)
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import InputFile

api = API(token=Token.from_env())
bot = Telegrinder(api)
wm = WaiterMachine(bot.dispatch)
kitten_pic = InputFile.from_path(pathlib.Path("examples/assets/kitten.jpg"))
logger.set_level("DEBUG")

bot.dispatch.message.auto_rules = IsPrivate() & IsUser()


class DummyMiddleware(ABCMiddleware):
    def pre(self) -> bool:
        return True

    def post(self) -> None:
        return None


@bot.on.message(Text("/start"))
async def start(message: Message, me: Me):
    await message.answer(
        "Hello, {}! It's {}. How are you today?".format(
            message.from_user.first_name,
            me.first_name,
        ),
    )
    m, _ = await wm.wait(
        hasher=MESSAGE_FROM_USER_IN_CHAT((message.from_user.id, message.chat_id)),
        release=Text(["fine", "bad"], ignore_case=True),
        on_miss=MessageReplyHandler("Fine or bad", as_reply=True),
    )

    match m.text.unwrap().lower():
        case "fine":
            await m.reply("Cool!")
        case "bad":
            await message.answer_photo(
                photo=kitten_pic,
                caption="I'm sorry... You prob need some kitten pictures",
            )


@bot.on.message(Text("/react"))
async def react(message: Message, context: Context):
    await message.reply("Send me any message...")
    msg, _ = await wm.wait(
        hasher=MESSAGE_FROM_USER_IN_CHAT((message.from_user.id, message.chat_id)),
        release=HasText(),
        on_miss=MessageReplyHandler("Your message has no text!"),
        lifespan=DummyMiddleware().to_lifespan(message, context),
    )
    await msg.react("🔥")


@bot.on.message(Markup("/reverse <text>"))
async def reverse(message: Message, text: str):
    await message.answer(
        text=f"Okay, its.. {text[-1].upper()}.. {text[-2].upper()}.. {text[::-1].lower().capitalize()}",
    )
    if text[::-1].lower().replace(" ", "") == text.lower().replace(" ", ""):
        await message.answer(
            text="Wow.. Seems like this is a palindrome.",
        )


# In order to make argument optional use this trick:
#   1) declare patterns ordered by their broadness
#   2) don't forget to set default value for optional arguments
@bot.on.message(Markup(["/predict", "/predict <thing>"]))
async def predict(thing: str = "it"):
    probability_percent = random.randint(0, 100)
    return f"I predict the probability {thing} will happen is {probability_percent}%"


@bot.on.message(FuzzyText("hello"))
async def hello(message: Message):
    await message.reply("Hi!")


@bot.on.message(FuzzyText("freeze"))
async def freeze_handler(message: Message):
    msg = (
        await message.answer(
            text="Well ok freezing",
            reply_markup=InlineKeyboard().add(InlineButton("Unfreeze", callback_data="unfreeze")).get_markup(),
        )
    ).unwrap()

    with bot.on.global_middleware.apply_filters(
        source_filter=(as_node(UserId), message.from_user.id, IsUpdateType(UpdateType.CALLBACK_QUERY)),
    ):
        cb, _ = await wm.wait(
            hasher=CALLBACK_QUERY_FOR_MESSAGE(msg.message_id),
            release=CallbackDataEq("unfreeze"),
        )
        await cb.edit_text("Wow heated!")


bot.run_forever(skip_updates=True)



import typing

from examples.nodes import DB, create_tables
from telegrinder import API, Message, Telegrinder, Token, node
from telegrinder.bot.dispatch import Context
from telegrinder.bot.rules import ABCRule, Markup, Text
from telegrinder.modules import logger
from telegrinder.node import ChatSource, TextLiteral, scalar_node

MessageId = type("MessageId", (int,), {})

api = API(token=Token.from_env())
bot = Telegrinder(api)
logger.set_level("DEBUG")


class IsChat(ABCRule):
    async def check(self, chat: ChatSource) -> bool:
        return chat.id < 0


class IsAdmin(ABCRule):
    async def check(self, source: node.Source, db: DB, context: Context) -> bool:
        result = await db.execute("select * from admins where telegram_id = ?", (source.from_user.id,))  # type: ignore
        context["is_admin"] = True
        return bool(await result.fetchone())


@scalar_node
class IncomingMessageId:
    @classmethod
    def compose(cls, message: Message) -> MessageId:
        return MessageId(message.message_id)


async def promote(user_id: int, *, db: DB) -> None:
    await db.execute("insert into admins(telegram_id) values (?) on conflict do nothing", (user_id,))  # type: ignore
    await db.commit()  # type: ignore


@bot.on.message(IsChat())
async def photo_in_chat_handler(message: Message, p: node.Photo) -> None:
    photo_size = p.sizes[-1]
    await message.answer("Photo ratio H/W: {}".format(photo_size.height / photo_size.width))


@bot.on.message(Text("/message_id"))
async def reply_handler(_: Message, message_id: IncomingMessageId) -> str:
    return f"Your message id: {message_id}"


@bot.on.message()
async def handle_texts(texts: TextLiteral["hello", "hi", "hilo"]) -> typing.Literal["hilo", "hello", "hi"]:
    match texts:
        case "hello":
            return "hilo"
        case "hi":
            return "hello"
        case "hilo":
            return "hi"


### Two handlers below require DB node
# DB node is marked with scope = PER_EVENT
# therefore DB node will only be resolved once on event
# If you run this code you will see that connection opening
# and closing is going to happen only once.
@bot.on.message()
async def photo_handler(
    message: Message,
    db: DB,
    photo: node.Photo,
) -> None:
    await message.answer("Got a photo in private message")


@bot.on.message()
async def integer_handler(
    message: Message,
    db: DB,
    i: node.TextInteger,
) -> None:
    await message.answer(f"{i} + 3 = {i + 3}")


@bot.on.message(IsAdmin(), Text("/op"))
async def add_admin_handler(message: Message, db: DB) -> str | None:
    if not message.reply_to_message:
        return "Need reply"
    await promote(message.reply_to_message.unwrap().from_user.id, db=db)
    await message.answer("Done")


@bot.on.message(Markup("/getadmin <token>"))
async def getadmin_handler(message: Message, token: str, db: DB) -> str:
    if token != api.token:
        return "Wrong token"
    await promote(message.from_user.id, db=db)
    return "Done"


@bot.on.message(Text("/amiadmin"), IsAdmin().as_optional())
async def amiadmin_handler(message: Message, is_admin: bool = False):
    await message.answer("You are " + ("not " if not is_admin else "") + "an admin")


bot.loop_wrapper.lifespan.on_startup(create_tables())
bot.run_forever()



import typing

from telegrinder.api.api import API
from telegrinder.bot.rules.abc import ABCRule, Context, check_rule
from telegrinder.types.objects import Update


class If(ABCRule):
    def __init__(self, condition: ABCRule) -> None:
        self.conditions = [condition]

    async def check(self, update: Update, api: API, ctx: Context) -> bool:
        for condition in self.conditions[:-1]:
            if not await check_rule(api, condition, update, ctx):
                return True

        return await check_rule(api, self.conditions[-1], update, ctx)

    def then(self, condition: ABCRule, /) -> typing.Self:
        self.conditions.append(condition)
        return self


__all__ = ("If",)



from telegrinder.tools.aio import (
    cancel_future,
    maybe_awaitable,
    next_generator,
    run_task,
    send_generator_value,
    stop_generator,
)
from telegrinder.tools.callback_data_serialization import (
    ABCDataSerializer,
    JSONSerializer,
    MsgPackSerializer,
)
from telegrinder.tools.formatting import (
    FormatString,
    HTMLFormatter,
    block_quote,
    bold,
    code_inline,
    escape,
    italic,
    link,
    mention,
    pre_code,
    spoiler,
    strike,
    tg_bot_attach_open_any_chat,
    tg_bot_attach_open_current_chat,
    tg_bot_attach_open_specific_chat,
    tg_bot_start_link,
    tg_bot_startchannel_link,
    tg_bot_startgroup_link,
    tg_chat_folder_link,
    tg_chat_invite_link,
    tg_direct_mini_app_link,
    tg_emoji,
    tg_emoji_link,
    tg_emoji_stickerset_link,
    tg_invoice_link,
    tg_language_pack_link,
    tg_main_mini_app_link,
    tg_mention_link,
    tg_open_message_link,
    tg_premium_multigift_link,
    tg_premium_offer_link,
    tg_private_channel_boost_link,
    tg_private_message_link,
    tg_public_channel_boost_link,
    tg_public_message_link,
    tg_public_username_link,
    tg_share_link,
    tg_story_link,
    underline,
)
from telegrinder.tools.fullname import fullname
from telegrinder.tools.global_context import (
    ABCGlobalContext,
    CtxVar,
    GlobalContext,
    GlobalCtxVar,
    TelegrinderContext,
    ctx_var,
    runtime_init,
)
from telegrinder.tools.input_file_directory import InputFileDirectory
from telegrinder.tools.lifespan import Lifespan
from telegrinder.tools.limited_dict import LimitedDict
from telegrinder.tools.loop_wrapper import DelayedTask, LoopWrapper
from telegrinder.tools.magic import (
    Annotations,
    Bundle,
    bundle,
    get_default_args,
    get_func_annotations,
    get_func_parameters,
    get_generic_parameters,
    join_dicts,
    resolve_arg_names,
    resolve_kwonly_arg_names,
    resolve_posonly_arg_names,
    shortcut,
)
from telegrinder.tools.parse_mode import ParseMode
from telegrinder.tools.singleton import ABCSingleton, ABCSingletonMeta, Singleton, SingletonMeta

__all__ = (
    "ABCDataSerializer",
    "ABCGlobalContext",
    "ABCSingleton",
    "ABCSingletonMeta",
    "Annotations",
    "Bundle",
    "CtxVar",
    "DelayedTask",
    "FormatString",
    "GlobalContext",
    "GlobalCtxVar",
    "HTMLFormatter",
    "InputFileDirectory",
    "JSONSerializer",
    "Lifespan",
    "LimitedDict",
    "LoopWrapper",
    "MsgPackSerializer",
    "ParseMode",
    "Singleton",
    "SingletonMeta",
    "TelegrinderContext",
    "block_quote",
    "bold",
    "bundle",
    "cancel_future",
    "code_inline",
    "ctx_var",
    "escape",
    "fullname",
    "get_default_args",
    "get_func_annotations",
    "get_func_parameters",
    "get_generic_parameters",
    "italic",
    "join_dicts",
    "link",
    "maybe_awaitable",
    "mention",
    "next_generator",
    "pre_code",
    "resolve_arg_names",
    "resolve_kwonly_arg_names",
    "resolve_posonly_arg_names",
    "run_task",
    "runtime_init",
    "send_generator_value",
    "shortcut",
    "spoiler",
    "stop_generator",
    "strike",
    "tg_bot_attach_open_any_chat",
    "tg_bot_attach_open_current_chat",
    "tg_bot_attach_open_specific_chat",
    "tg_bot_start_link",
    "tg_bot_startchannel_link",
    "tg_bot_startgroup_link",
    "tg_chat_folder_link",
    "tg_chat_invite_link",
    "tg_direct_mini_app_link",
    "tg_emoji",
    "tg_emoji_link",
    "tg_emoji_stickerset_link",
    "tg_invoice_link",
    "tg_language_pack_link",
    "tg_main_mini_app_link",
    "tg_mention_link",
    "tg_open_message_link",
    "tg_premium_multigift_link",
    "tg_premium_offer_link",
    "tg_private_channel_boost_link",
    "tg_private_message_link",
    "tg_public_channel_boost_link",
    "tg_public_message_link",
    "tg_public_username_link",
    "tg_share_link",
    "tg_story_link",
    "underline",
)



from telegrinder.tools.keyboard.buttons.base import BaseButton, BaseStaticButton
from telegrinder.tools.keyboard.buttons.buttons import Button, InlineButton
from telegrinder.tools.keyboard.buttons.static_buttons import StaticButton, StaticInlineButton
from telegrinder.tools.keyboard.buttons.tools import RowButtons

__all__ = (
    "BaseButton",
    "BaseStaticButton",
    "Button",
    "InlineButton",
    "RowButtons",
    "StaticButton",
    "StaticInlineButton",
)



import pytest

from telegrinder.api.api import API, Token
from telegrinder.api.error import APIError
from telegrinder.api.response import APIResponse
from telegrinder.msgspec_utils import decoder
from telegrinder.types.objects import User

from .test_utils import with_mocked_api

API_ERROR_RESPONSE = {"ok": False, "error_code": 404, "description": "Not Found"}

GET_ME_RAW_RESPONSE = """
{
  "ok": true,
  "result": {
    "id": 55732494,
    "is_bot": true,
    "first_name": "Cute",
    "username": "cute123_bot",
    "can_join_groups": true,
    "can_read_all_group_messages": false,
    "supports_inline_queries": true
  }
}
"""
GET_ME_USER_DICT = decoder.decode(GET_ME_RAW_RESPONSE, type=dict)
GET_ME_API_RESPONSE = decoder.decode(GET_ME_RAW_RESPONSE, type=APIResponse).to_result()
GET_ME_USER_MODEL = decoder.decode(GET_ME_API_RESPONSE.unwrap(), type=User)


@pytest.mark.asyncio()
@with_mocked_api(GET_ME_RAW_RESPONSE)
async def test_get_me_raw(api: API):
    raw_me = await api.request_raw("getMe")

    assert raw_me
    assert raw_me == GET_ME_API_RESPONSE
    assert raw_me.unwrap() == GET_ME_API_RESPONSE.unwrap()


@pytest.mark.asyncio()
@with_mocked_api(GET_ME_RAW_RESPONSE)
async def test_get_me_method(api: API):
    me = await api.get_me()

    assert me
    assert me.unwrap() == GET_ME_USER_MODEL
    assert me.unwrap().id == 55732494
    assert me.unwrap().first_name == "Cute"
    assert me.unwrap().is_bot == True


@pytest.mark.asyncio()
@with_mocked_api(GET_ME_USER_DICT)
async def test_api_request_method(api: API):
    response = await api.request("getMe")

    assert response
    assert response.unwrap() == GET_ME_USER_DICT["result"]
    assert isinstance(response.value, dict)
    assert response.value["id"] == GET_ME_USER_DICT["result"]["id"]
    assert response.value["first_name"] == GET_ME_USER_DICT["result"]["first_name"]
    assert response.value["is_bot"] == GET_ME_USER_DICT["result"]["is_bot"]


@pytest.mark.asyncio()
@with_mocked_api(API_ERROR_RESPONSE)
async def test_api_error(api: API):
    response = await api.request("someMethod")

    assert not response
    assert isinstance(response.error, APIError)
    assert response.error.code == 404
    assert response.error.error == "Not Found"
    with pytest.raises(APIError, match="Not Found"):
        response.unwrap()


@pytest.mark.asyncio()
@with_mocked_api(None)
async def test_api_id(api: API):
    assert api.id == 123


@pytest.mark.asyncio()
@with_mocked_api(None)
async def test_api_token(api: API):
    assert api.token == Token("123:ABCdef")


@pytest.mark.asyncio()
@with_mocked_api(b"somefile")
async def test_download_file(api: API):
    assert await api.download_file("cool_file") == b"somefile"



import datetime as dt
import typing
from contextlib import contextmanager

import msgspec
from fntypes.option import Nothing, Some
from fntypes.result import Error, Ok, Result
from fntypes.variative import Variative

from telegrinder.msgspec_utils.abc import SupportsCast
from telegrinder.msgspec_utils.custom_types.datetime import datetime, timedelta
from telegrinder.msgspec_utils.custom_types.enum_meta import BaseEnumMeta
from telegrinder.msgspec_utils.tools import bundle, fullname, get_origin

type Context = dict[str, typing.Any]
type Order = typing.Literal["deterministic", "sorted"]
type EncHook[T] = typing.Callable[typing.Concatenate[T, ...], typing.Any]


def to_builtins(
    obj: typing.Any,
    *,
    str_keys: bool = False,
    builtin_types: typing.Iterable[type[typing.Any]] | None = None,
    order: Order | None = None,
    context: Context | None = None,
) -> Result[typing.Any, msgspec.ValidationError]:
    try:
        return Ok(
            encoder.to_builtins(
                obj,
                str_keys=str_keys,
                builtin_types=builtin_types,
                order=order,
                context=context,
            ),
        )
    except msgspec.ValidationError as error:
        return Error(error)


class Encoder:
    """Class `Encoder` for `msgspec` module with encode hooks for objects.

    ```
    from datetime import datetime

    class MyDatetime(datetime):
        ...

    encoder = Encoder()
    encoder.enc_hooks[MyDatetime] = lambda d: int(d.timestamp())

    encoder.enc_hook(MyDatetime.now())  #> 1713354732
    encoder.encode({"digit": Digit.ONE})  #> '{"digit":1}'
    ```
    """

    cast_types: dict[type[typing.Any], type[SupportsCast]]
    enc_hooks: dict[typing.Any, EncHook[typing.Any]]
    abstract_enc_hooks: dict[typing.Any, EncHook[typing.Any]]

    def __init__(self) -> None:
        self.cast_types = {  # type: ignore
            dt.datetime: datetime,
            dt.timedelta: timedelta,
        }
        self.enc_hooks = {
            Some: lambda some: some.value,
            Nothing: lambda _: None,
            Variative: lambda variative: variative.v,
            datetime: lambda date: int(date.timestamp()),
            timedelta: lambda time: time.total_seconds(),
        }
        self.abstract_enc_hooks = {
            BaseEnumMeta: lambda enum_member: enum_member.value,
        }

    def __repr__(self) -> str:
        return "<{}: cast_types={!r}, enc_hooks={!r}, abstract_enc_hooks={!r}>".format(
            type(self).__name__,
            self.cast_types,
            self.enc_hooks,
            self.abstract_enc_hooks,
        )

    @contextmanager
    def __call__(
        self,
        *,
        decimal_format: typing.Literal["string", "number"] = "string",
        uuid_format: typing.Literal["canonical", "hex"] = "canonical",
        order: Order | None = None,
        context: Context | None = None,
    ) -> typing.Generator[msgspec.json.Encoder, typing.Any, None]:
        """Context manager returns the `msgspec.json.Encoder` object with passed the `enc_hook`."""
        yield msgspec.json.Encoder(
            enc_hook=self.enc_hook(context),
            decimal_format=decimal_format,
            uuid_format=uuid_format,
            order=order,
        )

    def add_enc_hook[T](self, t: type[T], /) -> typing.Callable[[EncHook[T]], EncHook[T]]:
        def decorator(func: EncHook[T], /) -> EncHook[T]:
            encode_hook = self.enc_hooks.setdefault(get_origin(t), func)
            return func if encode_hook is not func else encode_hook

        return decorator

    def add_abstract_enc_hook[T](self, abstract_type: type[T], /) -> typing.Callable[[EncHook[T]], EncHook[T]]:
        def decorator(func: EncHook[T], /) -> EncHook[T]:
            return self.abstract_enc_hooks.setdefault(get_origin(abstract_type), func)

        return decorator

    def get_abstract_enc_hook(self, subtype: type[typing.Any], /) -> EncHook[typing.Any] | None:
        for abstract, enc_hook in self.abstract_enc_hooks.items():
            if issubclass(subtype, abstract) or issubclass(type(subtype), abstract):
                return enc_hook

        return None

    def enc_hook(self, context: Context | None = None, /) -> EncHook[typing.Any]:
        def inner(obj: typing.Any, /) -> typing.Any:
            origin_type = get_origin(obj)

            if (enc_hook_func := self.enc_hooks.get(origin_type)) is None and (
                enc_hook_func := self.get_abstract_enc_hook(origin_type)
            ) is None:
                raise NotImplementedError(
                    f"Not implemented encode hook for object of type `{fullname(origin_type)}`. "
                    "You can implement encode hook for this object.",
                )

            return bundle(enc_hook_func, context or {}, start_idx=1)(obj)

        return inner

    @typing.overload
    def encode(
        self,
        obj: typing.Any,
        *,
        order: Order | None = None,
        context: Context | None = None,
    ) -> str: ...

    @typing.overload
    def encode(
        self,
        obj: typing.Any,
        *,
        as_str: typing.Literal[True],
        order: Order | None = None,
        context: Context | None = None,
    ) -> str: ...

    @typing.overload
    def encode(
        self,
        obj: typing.Any,
        *,
        as_str: typing.Literal[False],
        order: Order | None = None,
        context: Context | None = None,
    ) -> bytes: ...

    def encode(
        self,
        obj: typing.Any,
        *,
        as_str: bool = True,
        order: Order | None = None,
        context: Context | None = None,
    ) -> str | bytes:
        buf = msgspec.json.encode(obj, enc_hook=self.enc_hook(context), order=order)
        return buf.decode() if as_str else buf

    def to_builtins(
        self,
        obj: typing.Any,
        *,
        str_keys: bool = False,
        builtin_types: typing.Iterable[type[typing.Any]] | None = None,
        order: Order | None = None,
        context: Context | None = None,
    ) -> typing.Any:
        return msgspec.to_builtins(
            obj,
            str_keys=str_keys,
            builtin_types=builtin_types,
            enc_hook=self.enc_hook(context),
            order=order,
        )

    def cast(self, obj: typing.Any, /) -> typing.Any:
        if (caster := self.cast_types.get(get_origin(obj))) is not None:
            return caster.cast(obj)
        return obj


encoder: typing.Final[Encoder] = Encoder()


__all__ = ("Encoder", "encoder", "to_builtins")



from fntypes.result import Ok

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.node.base import IsNode, is_node
from telegrinder.node.composer import compose_nodes
from telegrinder.types.objects import Update


class NodeRule(ABCRule):
    def __init__(self, *nodes: IsNode | tuple[str, IsNode]) -> None:
        self.nodes: list[IsNode] = []
        self.node_keys: list[str | None] = []

        for binding in nodes:
            node_key, node_t = binding if isinstance(binding, tuple) else (None, binding)
            if not is_node(node_t):
                continue
            self.nodes.append(node_t)
            self.node_keys.append(node_key)

    async def check(self, update: Update, api: API, context: Context) -> bool:
        result = await compose_nodes(
            nodes={f"node_{i}": node for i, node in enumerate(self.nodes)},
            ctx=context,
            data={Update: update, API: api},
        )

        match result:
            case Ok(collection):
                resolved_nodes = collection.sessions
            case _:
                return False

        for i, (node_key, node_value) in enumerate(resolved_nodes.items()):
            if (key := self.node_keys[i]) and node_key == key:
                context[key] = node_value

        await collection.close_all()
        return True


__all__ = ("NodeRule",)



from __future__ import annotations

import dataclasses
import typing

from telegrinder.tools.keyboard.abc import DictStrAny
from telegrinder.tools.keyboard.base import BaseStaticKeyboard
from telegrinder.tools.keyboard.data import KeyboardModel
from telegrinder.types.objects import InlineKeyboardMarkup, ReplyKeyboardMarkup, ReplyKeyboardRemove

type RawKeyboard = typing.Iterable[typing.Iterable[DictStrAny]]


def create_keyboard(cls_: type[BaseStaticKeyboard], /) -> RawKeyboard:
    buttons = cls_.get_buttons()
    max_in_row = cls_.__max_in_row__
    keyboard = [[]]

    for button in buttons.values():
        keyboard[-1].append(button.get_data())

        if (button.row is True or len(keyboard[-1]) >= max_in_row) and keyboard[-1]:
            keyboard.append([])

    return tuple(tuple(x) for x in keyboard if x)


class StaticKeyboard(BaseStaticKeyboard):
    __keyboard__: KeyboardModel

    def __init_subclass__(
        cls,
        *,
        resize_keyboard: bool = True,
        one_time_keyboard: bool = False,
        selective: bool = False,
        is_persistent: bool = False,
        input_field_placeholder: str | None = None,
        max_in_row: int = 3,
    ) -> None:
        cls.__max_in_row__ = max_in_row
        cls.__keyboard__ = KeyboardModel(
            keyboard=create_keyboard(cls),
            resize_keyboard=resize_keyboard,
            one_time_keyboard=one_time_keyboard,
            selective=selective,
            is_persistent=is_persistent,
            input_field_placeholder=input_field_placeholder,
        )

    @classmethod
    def dict(cls) -> DictStrAny:
        return dataclasses.asdict(cls.__keyboard__)

    @classmethod
    def get_markup(cls) -> ReplyKeyboardMarkup:
        return ReplyKeyboardMarkup.from_dict(cls.dict())

    @classmethod
    def get_keyboard_remove(cls) -> ReplyKeyboardRemove:
        return ReplyKeyboardRemove.from_data(
            remove_keyboard=True,
            selective=cls.__keyboard__.selective,
        )


class StaticInlineKeyboard(BaseStaticKeyboard):
    __keyboard__: RawKeyboard

    def __init_subclass__(cls, *, max_in_row: int = 3) -> None:
        cls.__max_in_row__ = max_in_row
        cls.__keyboard__ = create_keyboard(cls)

    @classmethod
    def dict(cls) -> DictStrAny:
        return dict(inline_keyboard=cls.__keyboard__)

    @classmethod
    def get_markup(cls) -> InlineKeyboardMarkup:
        return InlineKeyboardMarkup.from_dict(cls.dict())


__all__ = ("StaticInlineKeyboard", "StaticKeyboard")



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Markup, RuleEnum, Text

api = API(token=Token.from_env())
bot = Telegrinder(api)
logger.set_level("INFO")


class CancelOrUsername(RuleEnum):
    CANCEL = Text("/cancel")
    USERNAME = Markup("@<username>") | Markup("t.me/<username>")


@bot.on.message(CancelOrUsername.USERNAME)
async def handle_username(m: Message, username: str | None = None):
    await m.answer(f"Username: {username}")


@bot.on.message(CancelOrUsername())
async def handle_cancel(m: Message):
    await m.answer("Uoops everything was canceled so sad... I wish it could continue")


bot.run_forever()



import typing

from telegrinder.types.methods_utils import get_params
from telegrinder.types.objects import (
    InputFile,
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
    MessageEntity,
    ReactionEmoji,
    ReactionType,
    ReactionTypeEmoji,
)

type InputMedia = typing.Union[
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
]

INPUT_MEDIA_TYPES: typing.Final[dict[str, type[InputMedia]]] = {
    "animation": InputMediaAnimation,
    "audio": InputMediaAudio,
    "document": InputMediaDocument,
    "photo": InputMediaPhoto,
    "video": InputMediaVideo,
}


def compose_reactions(
    reactions: str | ReactionEmoji | ReactionType | list[str | ReactionEmoji | ReactionType],
    /,
) -> list[ReactionType]:
    if not isinstance(reactions, list):
        reactions = [reactions]
    return [
        (
            ReactionTypeEmoji(emoji)
            if isinstance(emoji, ReactionEmoji)
            else (ReactionTypeEmoji(ReactionEmoji(emoji)) if isinstance(emoji, str) else emoji)
        )
        for emoji in reactions
    ]


def input_media(
    type: typing.Literal["animation", "audio", "document", "photo", "video"],
    media: str | InputFile,
    *,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    **other: typing.Any,
) -> InputMedia:
    return INPUT_MEDIA_TYPES[type](**get_params(locals()))


__all__ = ("compose_reactions", "input_media")



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Text

SECRET_CHAT_ID = 123456789
bot = Telegrinder(API(Token.from_env()))
logger.set_level("INFO")


@bot.loop_wrapper.timer(minutes=1)
async def once_yum():
    print("yum-yum!")


@bot.loop_wrapper.interval(seconds=10)
async def repeat_yum():
    print("repeat yum-yum!!!")


@bot.loop_wrapper.lifespan.on_startup
async def hello():
    await bot.api.send_message(
        chat_id=SECRET_CHAT_ID,
        text="Hello!!!",
    )


@bot.loop_wrapper.lifespan.on_shutdown
async def bye():
    print("Bye!!!")


@bot.on.message(Text("/hello"))
async def hello_handler(m: Message):
    await m.reply("Hello, world!")


bot.run_forever()



import typing
from contextlib import suppress
from functools import cached_property

import msgspec

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.markup import Markup, PatternLike, check_string
from telegrinder.msgspec_utils.json import loads
from telegrinder.node.base import Node
from telegrinder.node.payload import Payload, PayloadData
from telegrinder.tools.callback_data_serialization.abc import ABCDataSerializer, ModelType
from telegrinder.tools.callback_data_serialization.json_ser import JSONSerializer


class PayloadRule[Data](ABCRule):
    def __init__(
        self,
        data_type: type[Data],
        serializer: type[ABCDataSerializer[Data]],
        *,
        alias: str | None = None,
    ) -> None:
        self.data_type = data_type
        self.serializer = serializer
        self.alias = alias or "data"

    @cached_property
    def required_nodes(self) -> dict[str, type[Node]]:
        return {"payload": PayloadData[self.data_type, self.serializer]}  # type: ignore

    def check(self, payload: PayloadData[Data], context: Context) -> typing.Literal[True]:
        context.set(self.alias, payload)
        return True


class PayloadModelRule[Model: ModelType](PayloadRule):
    def __init__(
        self,
        model_t: type[Model],
        /,
        *,
        serializer: type[ABCDataSerializer[Model]] | None = None,
        alias: str | None = None,
    ) -> None:
        super().__init__(model_t, serializer or JSONSerializer, alias=alias or "model")


class PayloadEqRule(ABCRule):
    def __init__(self, payloads: str | list[str], /) -> None:
        self.payloads = [payloads] if isinstance(payloads, str) else payloads

    def check(self, payload: Payload) -> bool:
        return any(p == payload for p in self.payloads)


class PayloadMarkupRule(ABCRule):
    def __init__(self, pattern: PatternLike | list[PatternLike], /) -> None:
        self.patterns = Markup(pattern).patterns

    def check(self, payload: Payload, context: Context) -> bool:
        return check_string(self.patterns, payload, context)


class PayloadJsonEqRule(ABCRule):
    def __init__(self, payload: dict[str, typing.Any], /) -> None:
        self.payload = payload

    def check(self, payload: Payload) -> bool:
        with suppress(msgspec.DecodeError, msgspec.ValidationError):
            return self.payload == loads(payload)
        return False


__all__ = (
    "PayloadEqRule",
    "PayloadJsonEqRule",
    "PayloadMarkupRule",
    "PayloadModelRule",
    "PayloadRule",
)



import typing

import telegrinder.types as tg_types
from telegrinder.api.api import API
from telegrinder.node.attachment import Animation, Audio, Document, Photo, Video, VideoNote, Voice
from telegrinder.node.base import FactoryNode

type Attachment = Animation | Audio | Document | Photo | Video | VideoNote | Voice


class _FileId(FactoryNode):
    attachment_node: type[Attachment]

    def __class_getitem__(cls, attachment_node: type[Attachment], /):
        return cls(attachment_node=attachment_node)

    @classmethod
    def get_subnodes(cls):
        return {"attach": cls.attachment_node}

    @classmethod
    def compose(cls, attach: Attachment) -> str:
        if isinstance(attach, Photo):
            return attach.sizes[-1].file_id
        return attach.file_id


class _File(FactoryNode):
    attachment_node: type[Attachment]

    def __class_getitem__(cls, attachment_node: type[Attachment], /):
        return cls(attachment_node=attachment_node)

    @classmethod
    def get_subnodes(cls):
        return {"file_id": _FileId[cls.attachment_node]}

    @classmethod
    async def compose(cls, file_id: str, api: API) -> tg_types.File:
        return (await api.get_file(file_id=file_id)).expect("File can't be downloaded.")


if typing.TYPE_CHECKING:
    type FileId[T: Attachment] = str
    type File[T: Attachment] = tg_types.File
else:
    FileId = _FileId
    File = _File


__all__ = ("File", "FileId")



from telegrinder.api.api import API
from telegrinder.node.base import ComposeError, scalar_node
from telegrinder.node.scope import global_node
from telegrinder.types.objects import User


@scalar_node
@global_node
class Me:
    @classmethod
    async def compose(cls, api: API) -> User:
        me = await api.get_me()
        return me.expect(ComposeError("Can't complete api.get_me() request."))


__all__ = ("Me",)



from telegrinder.client.abc import ABCClient
from telegrinder.client.aiohttp import AiohttpClient
from telegrinder.client.form_data import MultipartFormProto, encode_form_data

__all__ = (
    "ABCClient",
    "AiohttpClient",
    "MultipartFormProto",
    "encode_form_data",
)



import typing

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.base import BaseReplyHandler
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.types.objects import InputFile


class PhotoReplyHandler(BaseReplyHandler):
    def __init__(
        self,
        photo: InputFile | str,
        *rules: ABCRule,
        caption: str | None = None,
        parse_mode: str | None = None,
        final: bool = True,
        as_reply: bool = False,
        preset_context: Context | None = None,
        **default_params: typing.Any,
    ) -> None:
        self.photo = photo
        self.parse_mode = parse_mode
        self.caption = caption
        super().__init__(
            *rules,
            final=final,
            as_reply=as_reply,
            preset_context=preset_context,
            **default_params,
        )

    async def handle(self, message: MessageCute) -> None:
        method = message.answer_photo if not self.as_reply else message.reply_photo
        await method(
            photo=self.photo,
            parse_mode=self.parse_mode,
            caption=self.caption,
            **self.default_params,
        )


__all__ = ("PhotoReplyHandler",)



from __future__ import annotations

import dataclasses
import types
import typing

import msgspec

from telegrinder.msgspec_utils import encoder
from telegrinder.tools.callback_data_serialization import ABCDataSerializer, JSONSerializer
from telegrinder.tools.keyboard.buttons.base import BaseButton
from telegrinder.types.objects import (
    CallbackGame,
    CopyTextButton,
    KeyboardButtonPollType,
    KeyboardButtonRequestChat,
    KeyboardButtonRequestUsers,
    LoginUrl,
    SwitchInlineQueryChosenChat,
    WebAppInfo,
)

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance

    from telegrinder.tools.keyboard.abc import ABCKeyboard
    from telegrinder.tools.keyboard.keyboard import InlineKeyboard, Keyboard

type CallbackData = str | bytes | dict[str, typing.Any] | DataclassInstance | msgspec.Struct


def _get_keyboard_class(name: str, /) -> type[ABCKeyboard]:
    from telegrinder.tools.keyboard.keyboard import InlineKeyboard, Keyboard  # noqa

    return locals()[name]


def _magic_keyboard(
    name_keyboard: str,
    button: BaseButton,
    other: typing.Any,
    /,
    *,
    row: bool = False,
) -> ABCKeyboard | types.NotImplementedType:
    keyboard_type: type[typing.Any] = _get_keyboard_class(name_keyboard)
    if not isinstance(other, keyboard_type | type(button)):
        return NotImplemented

    keyboard: ABCKeyboard = keyboard_type().add(button)
    if row:
        keyboard = keyboard.row()

    return keyboard.merge_to_last_row(other) if isinstance(other, keyboard_type) else keyboard.add(other)


class ConvertButtonMixin[T: ABCKeyboard]:
    if typing.TYPE_CHECKING:
        as_keyboard: type[T]
    else:
        KEYBOARD_CLASS_KEY = "__keyboard_class__"

        @classmethod
        def get_keyboard_class(cls):
            if (kb_cls := getattr(cls, cls.KEYBOARD_CLASS_KEY, None)) is not None:
                return kb_cls

            arg = None
            for base in cls.__orig_bases__:
                origin = typing.get_origin(base) or base
                if issubclass(origin, ConvertButtonMixin):
                    arg = typing.get_origin(arg := typing.get_args(base)[0]) or arg

            assert arg is not None
            kb_cls_name = arg.__forward_arg__ if isinstance(arg, typing.ForwardRef) else arg.__name__
            kb_cls = _get_keyboard_class(kb_cls_name)
            setattr(cls, cls.KEYBOARD_CLASS_KEY, kb_cls)
            return kb_cls

        def as_keyboard(self, *args, **kwargs):
            keyboard = self.get_keyboard_class()(*args, **kwargs)
            return keyboard.add(self)


@dataclasses.dataclass
class Button(BaseButton, ConvertButtonMixin["Keyboard"]):
    text: str
    request_contact: bool = dataclasses.field(default=False, kw_only=True)
    request_location: bool = dataclasses.field(default=False, kw_only=True)
    request_chat: KeyboardButtonRequestChat | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    request_user: KeyboardButtonRequestUsers | None = dataclasses.field(default=None, kw_only=True)
    request_poll: KeyboardButtonPollType | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    web_app: WebAppInfo | None = dataclasses.field(default=None, kw_only=True)

    def __and__(self, other: object, /) -> Keyboard:
        return _magic_keyboard("Keyboard", self, other)

    def __or__(self, other: object, /) -> Keyboard:
        return _magic_keyboard("Keyboard", self, other, row=True)


@dataclasses.dataclass
class InlineButton(BaseButton, ConvertButtonMixin["InlineKeyboard"]):
    text: str
    url: str | None = dataclasses.field(default=None, kw_only=True)
    login_url: LoginUrl | None = dataclasses.field(default=None, kw_only=True)
    pay: bool | None = dataclasses.field(default=None, kw_only=True)
    callback_data: CallbackData | None = dataclasses.field(default=None, kw_only=True)
    callback_data_serializer: dataclasses.InitVar[ABCDataSerializer[typing.Any] | None] = dataclasses.field(
        default=None,
        kw_only=True,
    )
    callback_game: CallbackGame | None = dataclasses.field(default=None, kw_only=True)
    copy_text: str | CopyTextButton | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query: str | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query_current_chat: str | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query_chosen_chat: SwitchInlineQueryChosenChat | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    web_app: str | WebAppInfo | None = dataclasses.field(default=None, kw_only=True)

    def __post_init__(self, callback_data_serializer: ABCDataSerializer[typing.Any] | None) -> None:
        if (
            callback_data_serializer is None
            and isinstance(self.callback_data, msgspec.Struct | dict)
            or dataclasses.is_dataclass(self.callback_data)
        ):
            callback_data_serializer = callback_data_serializer or JSONSerializer(
                type(self.callback_data),
            )

        if callback_data_serializer is not None:
            self.callback_data = callback_data_serializer.serialize(self.callback_data)
        elif self.callback_data is not None and not isinstance(self.callback_data, str | bytes):
            self.callback_data = encoder.encode(self.callback_data)

        if isinstance(self.copy_text, str):
            self.copy_text = CopyTextButton(text=self.copy_text)

        if isinstance(self.web_app, str):
            self.web_app = WebAppInfo(url=self.web_app)

    def __and__(self, other: object, /) -> InlineKeyboard:
        return _magic_keyboard("InlineKeyboard", self, other)

    def __or__(self, other: object, /) -> InlineKeyboard:
        return _magic_keyboard("InlineKeyboard", self, other, row=True)


__all__ = ("Button", "InlineButton")



import typing


class _LiteralMeta(type):
    __args__: tuple[typing.Any, ...]

    def __getitem__[T](cls: type[T], values: typing.Any, /) -> type[T]:
        values = (values,) if not isinstance(values, tuple) else values
        return _LiteralMeta(cls.__name__, (cls,), {"__args__": values})  # type: ignore

    def __instancecheck__(cls, instance: typing.Any, /) -> bool:
        return instance in cls.__args__


class _Literal(metaclass=_LiteralMeta):
    pass


if typing.TYPE_CHECKING:
    from typing import Literal
else:
    Literal = _Literal


__all__ = ("Literal",)



import asyncio
import dataclasses
import datetime
import typing

from telegrinder.modules import logger
from telegrinder.tools.aio import run_task
from telegrinder.tools.fullname import fullname

type CoroutineTask[T] = typing.Coroutine[typing.Any, typing.Any, T]
type CoroutineFunc[**P, T] = typing.Callable[P, CoroutineTask[T]]
type Task[**P, T] = CoroutineFunc[P, T] | CoroutineTask[T] | DelayedTask[typing.Callable[P, CoroutineTask[T]]]


def to_coroutine_task[**P, T](task: Task[P, T], /) -> CoroutineTask[T]:
    if asyncio.iscoroutinefunction(task) or isinstance(task, DelayedTask):
        task = task()
    elif not asyncio.iscoroutine(task):
        raise TypeError("Task should be coroutine or coroutine function.")
    return task


@dataclasses.dataclass
class DelayedTask[Function: CoroutineFunc[..., typing.Any]]:
    _cancelled: bool = dataclasses.field(default=False, init=False, repr=False)
    _task: asyncio.Task[typing.Any] | None = dataclasses.field(default=None, init=False, repr=False)

    function: Function
    seconds: float | datetime.timedelta
    repeat: bool = dataclasses.field(default=False, kw_only=True)

    def __post_init__(self) -> None:
        self.function.cancel = self.cancel

    @property
    def is_cancelled(self) -> bool:
        return self._cancelled

    @property
    def delay(self) -> float:
        return float(self.seconds) if isinstance(self.seconds, int | float) else self.seconds.total_seconds()

    async def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        self._task = self._task or asyncio.current_task()
        stopped = False

        while not self._cancelled and not stopped:
            await asyncio.sleep(self.delay)
            try:
                await self.function(*args, **kwargs)
            except BaseException:
                logger.exception(
                    "Delayed task for function `{}` caught an exception, traceback message below:",
                    fullname(self.function),
                )
            finally:
                stopped = not self.repeat

    def cancel(self) -> bool:
        if not self._cancelled:
            self._cancelled = True if self._task is None else self._task.cancel()
            self._task = None

        return self._cancelled


@dataclasses.dataclass(kw_only=True, slots=True, repr=False)
class Lifespan:
    _started: bool = dataclasses.field(default=False, init=False)
    startup_tasks: list[CoroutineTask[typing.Any]] = dataclasses.field(default_factory=lambda: [])
    shutdown_tasks: list[CoroutineTask[typing.Any]] = dataclasses.field(default_factory=lambda: [])

    def __repr__(self) -> str:
        return "<{}: started={}>".format(fullname(self), self._started)

    def __add__(self, other: object, /) -> typing.Self:
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.__class__(
            startup_tasks=self.startup_tasks + other.startup_tasks,
            shutdown_tasks=self.shutdown_tasks + other.shutdown_tasks,
        )

    def __iadd__(self, other: object, /) -> typing.Self:
        if not isinstance(other, self.__class__):
            return NotImplemented

        self.startup_tasks.extend(other.startup_tasks)
        self.shutdown_tasks.extend(other.shutdown_tasks)
        return self

    def __enter__(self) -> None:
        self.start()

    def __exit__(self, *args: typing.Any) -> None:
        self.shutdown()

    async def __aenter__(self) -> None:
        await self._start()

    async def __aexit__(self, *args: typing.Any) -> None:
        await self._shutdown()

    @property
    def started(self) -> bool:
        return self._started

    @staticmethod
    async def _run_tasks(tasks: list[CoroutineTask[typing.Any]], /) -> None:
        while tasks:
            await tasks.pop(0)

    async def _start(self) -> None:
        if not self._started:
            logger.debug("Running lifespan startup tasks")
            self._started = True
            await self._run_tasks(self.startup_tasks)

    async def _shutdown(self) -> None:
        if self._started:
            logger.debug("Running lifespan shutdown tasks")
            self._started = False
            await self._run_tasks(self.shutdown_tasks)

    def start(self) -> None:
        run_task(self._start())

    def shutdown(self) -> None:
        run_task(self._shutdown())

    def on_startup[**P, T](self, task: Task[P, T], /) -> Task[P, T]:
        self.startup_tasks.append(to_coroutine_task(task))
        return task

    def on_shutdown[**P, T](self, task: Task[P, T], /) -> Task[P, T]:
        self.shutdown_tasks.append(to_coroutine_task(task))
        return task


__all__ = ("CoroutineTask", "DelayedTask", "Lifespan", "run_task", "to_coroutine_task")



import dataclasses
import types
import typing
from abc import ABC, abstractmethod
from functools import cached_property

from fntypes.result import Error, Ok

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.modules import logger
from telegrinder.node.base import IsNode, get_nodes
from telegrinder.node.composer import compose_nodes
from telegrinder.tools import fullname
from telegrinder.tools.aio import maybe_awaitable
from telegrinder.tools.magic.function import bundle
from telegrinder.types.objects import Update

type ManagerFunction = typing.Callable[..., typing.Any | typing.Awaitable[typing.Any]]


def _get_types(x: typing.Any, /) -> type[typing.Any] | tuple[typing.Any, ...]:
    while True:
        if isinstance(x, types.UnionType | typing._UnionGenericAlias):  # type: ignore
            return tuple(_get_types(x) for x in typing.get_args(x))

        if isinstance(x, typing.TypeAliasType):
            x = x.__value__

        if isinstance(x, types.GenericAlias | typing._GenericAlias):  # type: ignore
            x = typing.get_origin(x)

        if isinstance(x, type):
            return x


def register_manager(return_type: typing.Any, /) -> typing.Callable[[ManagerFunction], "Manager"]:
    def wrapper(function: ManagerFunction, /) -> "Manager":
        function = function.__func__ if isinstance(function, classmethod | staticmethod) else function
        types = _get_types(return_type)
        return Manager((types,) if not isinstance(types, tuple) else types, function)

    return wrapper


@dataclasses.dataclass
class Manager:
    types: tuple[typing.Any, ...]
    function: ManagerFunction

    @cached_property
    def required_nodes(self) -> dict[str, IsNode]:
        return get_nodes(self.function, start_idx=1)

    async def __call__(
        self,
        response: typing.Any,
        update: Update,
        api: API,
        context: Context,
    ) -> None:
        data = {Update: update, API: api}
        node_col = None

        if self.required_nodes:
            match await compose_nodes(self.required_nodes, context, data=data):
                case Ok(value):
                    node_col = value
                case Error(compose_error):
                    logger.debug(
                        "Cannot compose nodes for return manager `{}`, error {!r}",
                        fullname(self.function),
                        compose_error.message,
                    )
                    return None

        temp_ctx = context.copy()
        try:
            bundle_function = bundle(self.function, {**data, Context: temp_ctx}, typebundle=True)
            bundle_function &= bundle(
                self.function,
                context | ({} if node_col is None else node_col.values),
            )
            await maybe_awaitable(bundle_function(response))
        finally:
            context |= temp_ctx

            if node_col is not None:
                await node_col.close_all()


class ABCReturnManager(ABC):
    @property
    @abstractmethod
    def managers(self) -> list[Manager]:
        pass

    @abstractmethod
    async def run(self, response: typing.Any, api: API, update: Update, context: Context) -> None:
        pass


class BaseReturnManager(ABCReturnManager):
    def __repr__(self) -> str:
        return "<{}: {}>".format(fullname(self), self.managers)

    @cached_property
    def managers(self) -> list[Manager]:
        return [
            manager
            for manager in (vars(BaseReturnManager) | vars(type(self))).values()
            if isinstance(manager, Manager)
        ]

    async def run(
        self,
        response: typing.Any,
        api: API,
        update: Update,
        context: Context,
    ) -> None:
        for manager in self.managers:
            if typing.Any in manager.types or type(response) in manager.types:
                logger.debug(
                    "Running manager `{}` for response `{!r}`",
                    fullname(manager.function),
                    response,
                )
                await manager(response, update, api, context)

    def register_manager(self, return_type: typing.Any, /) -> typing.Callable[[ManagerFunction], Manager]:
        def wrapper(function: ManagerFunction, /) -> Manager:
            manager = register_manager(return_type)(function)
            self.managers.append(manager)
            return manager

        return wrapper


__all__ = (
    "ABCReturnManager",
    "BaseReturnManager",
    "Manager",
    "register_manager",
)



import typing

if typing.TYPE_CHECKING:
    from telegrinder.tools.keyboard.buttons.base import BaseButton


class RowButtons[KeyboardButton: BaseButton]:
    buttons: typing.Iterable[KeyboardButton]

    def __init__(self, *buttons: KeyboardButton, auto_row: bool = True) -> None:
        self.buttons = buttons
        self.auto_row = auto_row

    def get_data(self) -> list[dict[str, typing.Any]]:
        return [b.get_data() for b in self.buttons]


__all__ = ("RowButtons",)



import typing

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.node.base import ComposeError, FactoryNode, scalar_node


@scalar_node
class Caption:
    @classmethod
    def compose(cls, message: MessageCute) -> str:
        if not message.caption:
            raise ComposeError("Message has no caption.")
        return message.caption.unwrap()


@scalar_node
class Text:
    @classmethod
    def compose(cls, message: MessageCute) -> str:
        if not message.text:
            raise ComposeError("Message has no text.")
        return message.text.unwrap()


@scalar_node
class TextInteger:
    @classmethod
    def compose(cls, text: Text | Caption) -> int:
        if not text.isdigit():
            raise ComposeError("Text is not digit.")
        return int(text)


if typing.TYPE_CHECKING:
    from typing import Literal as TextLiteral

else:

    class TextLiteral(FactoryNode):
        texts: tuple[str, ...]

        def __class_getitem__(cls, texts, /):
            return cls(texts=(texts,) if not isinstance(texts, tuple) else texts)

        @classmethod
        def compose(cls, text: Text) -> str:
            if text in cls.texts:
                return text
            raise ComposeError("Text mismatched literal.")


__all__ = ("Caption", "Text", "TextInteger", "TextLiteral")



from __future__ import annotations

import dataclasses
import inspect
import typing
from collections import OrderedDict
from functools import wraps

from fntypes.result import Result

from telegrinder.tools.magic.function import get_func_parameters
from telegrinder.types.methods_utils import get_params

type Executor[T] = typing.Callable[
    [T, str, dict[str, typing.Any]],
    typing.Awaitable[Result[typing.Any, APIError]],
]
type CuteMethod[T, **P, R] = typing.Callable[
    typing.Concatenate[T, P],
    typing.Awaitable[Result[R, APIError]],
]

if typing.TYPE_CHECKING:
    from telegrinder.api.error import APIError


class ShortcutMethod[T, **P, R](typing.Protocol):
    __name__: str
    __shortcut__: Shortcut[T]

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Result[R, APIError]: ...


@dataclasses.dataclass(slots=True, frozen=True)
class Shortcut[T]:
    method_name: str
    executor: Executor[T] | None = dataclasses.field(default=None, kw_only=True)
    custom_params: set[str] = dataclasses.field(default_factory=lambda: set[str](), kw_only=True)


@typing.overload
def shortcut[T, **P, R](
    method_name: str,
    /,
    *,
    custom_params: set[str] = ...,
) -> typing.Callable[[CuteMethod[T, P, R]], ShortcutMethod[T, P, R]]: ...


@typing.overload
def shortcut[T, **P, R](
    method_name: str,
    /,
    *,
    executor: Executor[T],
    custom_params: set[str] = ...,
) -> typing.Callable[[CuteMethod[T, P, R]], ShortcutMethod[T, P, R]]: ...


def shortcut[T, **P, R](
    method_name: str,
    /,
    *,
    executor: Executor[T] | None = None,
    custom_params: set[str] | None = None,
) -> typing.Callable[[CuteMethod[T, P, R]], ShortcutMethod[T, P, R]]:
    """Decorate a cute method as a shortcut."""

    def wrapper(func: CuteMethod[T, P, R]) -> ShortcutMethod[T, P, R]:
        @wraps(func)
        async def inner(
            self: T,
            *args: P.args,
            **kwargs: P.kwargs,
        ) -> Result[R, typing.Any]:
            if executor is None:
                return await func(self, *args, **kwargs)

            params: dict[str, typing.Any] = OrderedDict()
            func_params = get_func_parameters(func)

            for index, (arg, default) in enumerate(func_params["args"]):
                if len(args) > index:
                    params[arg] = args[index]
                elif default is not inspect.Parameter.empty:
                    params[arg] = default

            if var_args := func_params.get("var_star_args"):
                params[var_args] = args[len(func_params["args"]) :]

            for kwarg, default in func_params["kwargs"]:
                params[kwarg] = (
                    kwargs.pop(kwarg, default) if default is not inspect.Parameter.empty else kwargs.pop(kwarg)
                )

            if var_kwargs := func_params.get("var_star_kwargs"):
                params[var_kwargs] = kwargs.copy()

            return await executor(self, method_name, get_params(params))

        inner.__shortcut__ = Shortcut(  # type: ignore
            method_name=method_name,
            executor=executor,
            custom_params=custom_params or set(),
        )
        return inner  # type: ignore

    return wrapper


__all__ = ("Shortcut", "shortcut")



from telegrinder.api.api import API, compose_data, retryer
from telegrinder.api.error import APIError, APIServerError, InvalidTokenError
from telegrinder.api.response import APIResponse
from telegrinder.api.token import Token

__all__ = (
    "API",
    "APIError",
    "APIResponse",
    "APIServerError",
    "InvalidTokenError",
    "Token",
    "compose_data",
    "retryer",
)



from __future__ import annotations

import enum
import typing
from reprlib import recursive_repr

from fntypes import Some
from fntypes.option import Nothing, Option

if typing.TYPE_CHECKING:
    from telegrinder.api.api import API
    from telegrinder.bot.cute_types.update import UpdateCute
    from telegrinder.types.objects import Update

type Key = str | enum.Enum
type AnyValue = typing.Any


class Context(dict[str, AnyValue]):
    """Low level per event context storage."""

    update_cute: Option[UpdateCute] = Nothing()
    exception_update: Option[BaseException] = Nothing()

    def __init__(self, **kwargs: AnyValue) -> None:
        dict.__init__(self, **kwargs)

    @recursive_repr()
    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, ", ".join(f"{k}={v!r}" for k, v in self.items()))

    def __setitem__(self, __key: Key, __value: AnyValue) -> None:
        dict.__setitem__(self, self.key_to_str(__key), __value)

    def __getitem__(self, __key: Key) -> AnyValue:
        return dict.__getitem__(self, self.key_to_str(__key))

    def __delitem__(self, __key: Key) -> None:
        dict.__delitem__(self, self.key_to_str(__key))

    def __setattr__(self, __name: str, __value: AnyValue) -> None:
        self.__setitem__(__name, __value)

    def __getattribute__(self, __name: str) -> AnyValue:
        cls = type(self)

        if __name in cls.__annotations__:
            return self[__name] if __name in self else super().__getattribute__(__name)

        if __name in _CONTEXT_CLASS_ATTRS:
            return super().__getattribute__(__name)

        return self.__getitem__(__name)

    def __delattr__(self, __name: str) -> None:
        self.__delitem__(__name)

    def add_update_cute(self, update: Update, bound_api: API, /) -> typing.Self:
        from telegrinder.bot.cute_types.update import UpdateCute

        self.update_cute = Some(UpdateCute.from_update(update, bound_api))
        return self

    def add_exception_update(self, exception_update: BaseException, /) -> typing.Self:
        self.exception_update = Some(exception_update)
        return self

    @staticmethod
    def key_to_str(key: Key) -> str:
        return key if isinstance(key, str) else str(key.value)

    def copy(self) -> typing.Self:
        return self.__class__(**dict.copy(self))

    def set(self, key: Key, value: AnyValue) -> None:
        self[key] = value

    @typing.overload
    def get(self, key: Key) -> AnyValue | None: ...

    @typing.overload
    def get[T](self, key: Key, default: T) -> T | AnyValue: ...

    @typing.overload
    def get(self, key: Key, default: None = None) -> AnyValue | None: ...

    def get[T](self, key: Key, default: T | None = None) -> T | AnyValue | None:
        return dict.get(self, key, default)

    def get_or_set[T](self, key: Key, default: T) -> T:
        if key not in self:
            self.set(key, default)
        return self.get(key, default)

    def delete(self, key: Key) -> None:
        del self[key]


_CONTEXT_CLASS_ATTRS = frozenset(Context.__dict__ | dict.__dict__ | object.__dict__)


__all__ = ("Context",)



import typing

if typing.TYPE_CHECKING:
    from fntypes.option import Option
else:
    import fntypes
    import msgspec

    class OptionMeta(type):
        def __instancecheck__(cls, __instance):
            return isinstance(__instance, (fntypes.option.Some | fntypes.option.Nothing, msgspec.UnsetType))

    class Option[Value](metaclass=OptionMeta):
        pass


__all__ = ("Option",)



import typing
from abc import ABC, abstractmethod
from collections import deque

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.types.objects import Update


class ABCView(ABC):
    def __repr__(self) -> str:
        return "<{}>".format(type(self).__name__)

    @abstractmethod
    async def check(self, event: Update) -> bool:
        pass

    @abstractmethod
    async def process(
        self,
        event: Update,
        api: API,
        context: Context,
    ) -> bool:
        pass

    @abstractmethod
    def load(self, external: typing.Self, /) -> None:
        pass


class ABCEventRawView(ABCView, ABC):
    handlers: deque[ABCHandler]


__all__ = ("ABCEventRawView", "ABCView")



import typing

from fntypes.result import Result

from telegrinder.api.api import APIError
from telegrinder.bot.cute_types.base import BaseCute, compose_method_params, shortcut
from telegrinder.types.methods_utils import get_params
from telegrinder.types.objects import *


class InlineQueryCute(BaseCute[InlineQuery], InlineQuery, kw_only=True):
    @property
    def from_user(self) -> User:
        return self.from_

    @shortcut("answer_inline_query", custom_params={"results", "inline_query_id"})
    async def answer(
        self,
        results: InlineQueryResult | list[InlineQueryResult],
        *,
        button: InlineQueryResultsButton | None = None,
        cache_time: int | None = None,
        inline_query_id: str | None = None,
        is_personal: bool | None = None,
        next_offset: str | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.answer_inline_query()`, see the [documentation](https://core.telegram.org/bots/api#answerinlinequery)

        Use this method to send answers to an inline query. On success, True is returned.
        No more than 50 results per query are allowed."""
        params = compose_method_params(
            get_params(locals()),
            self,
            default_params={("inline_query_id", "id")},
        )
        params["results"] = [results] if not isinstance(results, list) else results
        return await self.ctx_api.answer_inline_query(**params)


__all__ = ("InlineQueryCute",)



import typing


class SingletonMeta(type):
    if not typing.TYPE_CHECKING:
        __instance = None

        def __call__(cls, *args, **kwargs):
            if cls.__instance is None:
                cls.__instance = super().__call__(*args, **kwargs)
            return cls.__instance


class Singleton(metaclass=SingletonMeta):
    pass


__all__ = ("Singleton", "SingletonMeta")



import typing

from telegrinder.bot.dispatch.return_manager.message import MessageReturnManager
from telegrinder.bot.dispatch.view.base import BaseView
from telegrinder.types.enums import UpdateType

type MessageUpdateType = typing.Literal[
    UpdateType.MESSAGE,
    UpdateType.BUSINESS_MESSAGE,
    UpdateType.CHANNEL_POST,
    UpdateType.EDITED_BUSINESS_MESSAGE,
    UpdateType.EDITED_CHANNEL_POST,
    UpdateType.EDITED_MESSAGE,
]


class MessageView(BaseView):
    def __init__(self, update_type: MessageUpdateType) -> None:
        super().__init__(update_type)
        self.return_manager = MessageReturnManager()


__all__ = ("MessageView",)



import builtins
import inspect
import os.path
import sys
import types

import typing_extensions as typing

type RoutineMethodType = (
    types.MethodType | types.MethodDescriptorType | types.MethodWrapperType | types.BuiltinMethodType
)
type RoutineDescriptorType = types.MethodDescriptorType | types.GetSetDescriptorType

_BUILTINS: typing.Final[frozenset[typing.Any]] = frozenset(
    x for name in dir(builtins) if getattr((x := getattr(builtins, name)), "__module__", None) == "builtins"
)
_CACHE_KEY: typing.Final[str] = "__fullname_cache__"


def _is_builtin(obj: typing.Any, /) -> bool:
    return inspect.isbuiltin(obj) or obj in _BUILTINS


def _is_routine_method(obj: typing.Any, /) -> typing.TypeIs[RoutineMethodType]:
    return inspect.isbuiltin(obj) or inspect.ismethod(obj) or inspect.ismethodwrapper(obj)


def _is_routine_descriptor(obj: typing.Any, /) -> typing.TypeIs[RoutineDescriptorType]:
    return inspect.ismethoddescriptor(obj) or inspect.isgetsetdescriptor(obj)


def _module_name(module: types.ModuleType, /) -> str:
    if (mod_name := module.__name__) != "__main__":
        return mod_name

    mod_package = module.__package__ or ""
    mod_file = module.__file__
    if mod_file is None:
        return mod_package

    mod_fname = os.path.basename(mod_file).removesuffix(".py")
    return mod_package if mod_package in ("__init__", "__main__") else mod_fname


def fullname(obj: object, /) -> str:
    """The full name (`__module__.__name__`) of the object."""

    if isinstance(obj, staticmethod | classmethod):
        obj = obj.__func__

    elif (
        not inspect.isroutine(obj)
        and not inspect.isgetsetdescriptor(obj)
        and not isinstance(obj, type)
        and not inspect.ismodule(obj)
    ):
        obj = type(obj)

    if (name := fullname.__dict__.setdefault(_CACHE_KEY, dict()).get(obj)) is not None:
        return name

    if inspect.ismodule(obj):
        fullname.__dict__[_CACHE_KEY][obj] = name = _module_name(obj)
        return name

    qualname = obj.__qualname__
    orig_obj = obj

    if _is_routine_method(obj) or _is_routine_descriptor(obj):
        obj_cls = obj.__objclass__ if _is_routine_descriptor(obj) else obj.__self__
        obj = type(obj_cls) if not isinstance(obj_cls, type) else obj_cls

    module: str = getattr(orig_obj, "__module__", obj.__module__)
    if _is_builtin(obj) and module != builtins.__name__:
        module = builtins.__name__
    else:
        module = _module_name(sys.modules[module])

    fullname.__dict__[_CACHE_KEY][orig_obj] = name = ".".join((module, qualname))
    return name


__all__ = ("fullname",)



from telegrinder import API, InlineButton, InlineKeyboard, Message, Telegrinder, Token
from telegrinder.rules import CallbackDataMap, Text

bot = Telegrinder(API(Token.from_env()))
kb = (
    InlineKeyboard()
    .add(InlineButton("Apple", callback_data={"item": "apple", "amount": 5}))
    .add(InlineButton("Banana", callback_data={"item": "banana", "amount": 15}))
    .add(InlineButton("Pineapple", callback_data={"item": "pineapple", "amount": 30}))
).get_markup()


@bot.on.message(Text("/start"))
async def start(message: Message):
    await message.answer("Hello, choose something", reply_markup=kb)


@bot.on.callback_query(CallbackDataMap({"item": str, "amount": lambda v: isinstance(v, int) and v <= 20}))
async def chose(item: str, amount: int) -> str:
    return f"You chose: {item=}, {amount=}"


bot.run_forever()



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import StartCommand

api = API(token=Token.from_env())
bot = Telegrinder(api)
logger.set_level("INFO")


@bot.on.message(StartCommand(lambda x: (int(x) if x.isdigit() else None)))
async def start_handler(message: Message, param: int | None) -> None:
    if param is None:
        await message.answer("You have no integer start query((")
        return
    await message.answer(
        "Ahah you integer start query is so funny, its {0} and {0}-42={1}".format(param, param - 42)
    )


@bot.on.message(StartCommand(param_required=True, alias="name"))
async def start_with_name(message: Message, name: str):
    await message.reply(f"Hello, {name!r}!")


bot.run_forever()



from telegrinder import API, ABCMiddleware, Message, Telegrinder, Token
from telegrinder.bot import Context
from telegrinder.modules import logger
from telegrinder.rules import IsChat, IsPrivate, IsUser, Text
from telegrinder.tools.global_context import GlobalContext

api = API(token=Token.from_env())
bot = Telegrinder(api)
global_ctx = GlobalContext(counter=dict[int, int]())  # Let's imagine a dummy counter

logger.set_level("INFO")


@bot.on.message.register_middleware
class CountMiddleware(ABCMiddleware):
    def pre(self, message: Message, context: Context) -> bool:
        counter = global_ctx.get_value("counter", dict[int, int]).unwrap()
        counter[message.chat.id] = context.count = counter.get(message.chat.id, 0) + 1
        return True


@bot.on.message(IsChat(), IsUser(), Text("/testme"))
async def testme_in_chat(m: Message) -> None:
    await m.reply("You are surely not a bot")


# The variable count will be passed to handler
# only if it is declared in handler function arguments
@bot.on.message(IsPrivate(), Text("/count"))
async def testme_private(count: int) -> str:
    return f"You wrote me {count} messages since my last reload"


bot.run_forever()



import typing
from functools import cached_property
from http import HTTPStatus

from fntypes.misc import from_optional
from fntypes.option import Option


class ReprErrorMixin:
    def __repr__(self) -> str:
        return f"{type(self).__name__}: {self}"


class APIError(ReprErrorMixin, Exception):
    def __init__(
        self,
        code: int,
        error: str,
        data: dict[str, typing.Any],
    ) -> None:
        self.code, self.error, self.parameters = code, error, data
        super().__init__(code, error, data)

    @cached_property
    def status_code(self) -> HTTPStatus:
        return HTTPStatus(self.code)

    @property
    def retry_after(self) -> Option[int]:
        return from_optional(self.parameters.get("retry_after"))

    @property
    def migrate_to_chat_id(self) -> Option[int]:
        return from_optional(self.parameters.get("migrate_to_chat_id"))

    def __str__(self) -> str:
        return f"[{self.code}] ({self.status_code.name}) {self.error}"


class APIServerError(ReprErrorMixin, Exception):
    def __init__(self, message: str, retry_after: int) -> None:
        self.message = message
        self.retry_after = retry_after
        super().__init__(message, retry_after)

    def __str__(self) -> str:
        return self.message


class InvalidTokenError(BaseException):
    pass


__all__ = ("APIError", "APIServerError", "InvalidTokenError")



import abc

from telegrinder.tools.singleton.singleton import SingletonMeta


class ABCSingletonMeta(abc.ABCMeta, SingletonMeta):
    pass


class ABCSingleton(metaclass=ABCSingletonMeta):
    pass


__all__ = ("ABCSingleton", "ABCSingletonMeta")



import enum
import typing


class BaseEnumMeta(enum.EnumMeta, type):
    if typing.TYPE_CHECKING:

        class BaseEnumMeta(enum.Enum):  # noqa
            NOT_SUPPORTED = enum.auto()

        NOT_SUPPORTED: typing.Literal[BaseEnumMeta.NOT_SUPPORTED]

    else:

        @staticmethod
        def _member_missing(cls, value):
            from telegrinder.modules import logger

            logger.warning(
                "Unsupported value {!r} for enum of type {!r}. Probably teleginder needs to be "
                "updated to support the latest version of Telegram Bot API.",
                value,
                cls,
            )
            return cls._member_map_["NOT_SUPPORTED"]

        def __new__(
            metacls,
            cls,
            bases,
            classdict,
            *,
            boundary=None,
            _simple=False,
            **kwds,
        ):
            classdict["NOT_SUPPORTED"] = "NOT_SUPPORTED" if any(x in bases for x in (str, enum.StrEnum)) else -1
            classdict["_missing_"] = classmethod(BaseEnumMeta._member_missing)
            new_type = super().__new__(metacls, cls, bases, classdict, boundary=boundary, _simple=_simple, **kwds)
            return new_type


__all__ = ("BaseEnumMeta",)



from __future__ import annotations

import dataclasses
import typing

if typing.TYPE_CHECKING:
    from telegrinder.tools.keyboard.abc import DictStrAny


@dataclasses.dataclass(kw_only=True, frozen=True)
class KeyboardModel:
    keyboard: typing.Iterable[typing.Iterable[DictStrAny]] = dataclasses.field(default_factory=lambda: [[]])
    resize_keyboard: bool = dataclasses.field(default=True)
    one_time_keyboard: bool = dataclasses.field(default=False)
    selective: bool = dataclasses.field(default=False)
    is_persistent: bool = dataclasses.field(default=False)
    input_field_placeholder: str | None = dataclasses.field(default=None)


__all__ = ("KeyboardModel",)



from telegrinder.tools.formatting.deep_links import (
    tg_bot_attach_open_any_chat,
    tg_bot_attach_open_current_chat,
    tg_bot_attach_open_specific_chat,
    tg_bot_start_link,
    tg_bot_startchannel_link,
    tg_bot_startgroup_link,
    tg_chat_folder_link,
    tg_chat_invite_link,
    tg_direct_mini_app_link,
    tg_emoji_link,
    tg_emoji_stickerset_link,
    tg_invoice_link,
    tg_language_pack_link,
    tg_main_mini_app_link,
    tg_mention_link,
    tg_open_message_link,
    tg_premium_multigift_link,
    tg_premium_offer_link,
    tg_private_channel_boost_link,
    tg_private_message_link,
    tg_public_channel_boost_link,
    tg_public_message_link,
    tg_public_username_link,
    tg_share_link,
    tg_story_link,
)
from telegrinder.tools.formatting.html_formatter import (
    FormatString,
    HTMLFormatter,
    block_quote,
    bold,
    code_inline,
    escape,
    italic,
    link,
    mention,
    pre_code,
    spoiler,
    strike,
    tg_emoji,
    underline,
)

__all__ = (
    "FormatString",
    "HTMLFormatter",
    "block_quote",
    "bold",
    "code_inline",
    "escape",
    "italic",
    "link",
    "mention",
    "pre_code",
    "spoiler",
    "strike",
    "tg_bot_attach_open_any_chat",
    "tg_bot_attach_open_current_chat",
    "tg_bot_attach_open_specific_chat",
    "tg_bot_start_link",
    "tg_bot_startchannel_link",
    "tg_bot_startgroup_link",
    "tg_chat_folder_link",
    "tg_chat_invite_link",
    "tg_direct_mini_app_link",
    "tg_emoji",
    "tg_emoji_link",
    "tg_emoji_stickerset_link",
    "tg_invoice_link",
    "tg_language_pack_link",
    "tg_main_mini_app_link",
    "tg_mention_link",
    "tg_open_message_link",
    "tg_premium_multigift_link",
    "tg_premium_offer_link",
    "tg_private_channel_boost_link",
    "tg_private_message_link",
    "tg_public_channel_boost_link",
    "tg_public_message_link",
    "tg_public_username_link",
    "tg_share_link",
    "tg_story_link",
    "underline",
)



"""Webhook Bot example. This example uses fastapi + uvicorn."""

import asyncio
import os
import secrets
from contextlib import asynccontextmanager

import uvicorn  # type: ignore
from bot import dp  # type: ignore
from fastapi import FastAPI, Request, Response  # type: ignore

from telegrinder import API, Token
from telegrinder.modules import logger
from telegrinder.tools.loop_wrapper import LoopWrapper
from telegrinder.types.objects import Update
from telegrinder.verification_utils import verify_webapp_request

token = Token.from_env()
api = API(token=token)
loop_wrapper = LoopWrapper().bind_loop(loop_factory=asyncio.new_event_loop)

HOST = os.environ["HOST"]  # > host, for example: https://domain.com
PORT = int(os.environ["PORT"])  # > port, can be either 443, 80, 88, or 8443.
WEBHOOK_PATH = os.environ["WEBHOOK_PATH"] + token  # > webhook path, for example: /bot/ + token
WEBHOOK_URL = HOST + WEBHOOK_PATH  # > host + webhook path
SECRET_TOKEN = secrets.token_urlsafe(64)  # > random secret token


@asynccontextmanager
async def lifespan(_):
    await api.set_webhook(url=WEBHOOK_URL, secret_token=SECRET_TOKEN, drop_pending_updates=True)
    yield
    await api.delete_webhook(drop_pending_updates=True)


app = FastAPI(lifespan=lifespan)  # type: ignore


@app.post(WEBHOOK_PATH, response_class=Response)  # type: ignore
async def webhook_bot(request: Request) -> Response:  # type: ignore
    if not verify_webapp_request(SECRET_TOKEN, request.headers):  # type: ignore
        return Response(status_code=403)  # type: ignore

    update = Update.from_raw(await request.body())  # type: ignore
    logger.debug(
        "Webhook received update (update_id={}, update_type={!r})",
        update.update_id,
        update.update_type,
    )
    await loop_wrapper.create_task(dp.feed(update, api))
    return Response(status_code=200)  # type: ignore


async def run_application(server: uvicorn.Server, /) -> None:  # type: ignore
    await server.serve()  # type: ignore


if __name__ == "__main__":
    config = uvicorn.Config(app, host="0.0.0.0", port=PORT, loop="none")  # type: ignore
    loop_wrapper.add_task(run_application(uvicorn.Server(config)))  # type: ignore
    loop_wrapper.run()



import dataclasses

import typing_extensions as typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.node.base import ComposeError, DataNode
from telegrinder.node.utility import TypeArgs

type ExceptionType = type[BaseException]

ExceptionT = typing.TypeVar("ExceptionT", bound=BaseException, default=BaseException)
ExceptionTs = typing.TypeVarTuple("ExceptionTs", default=typing.Unpack[tuple[BaseException, ...]])


def can_catch[ExceptionT: BaseException](
    exc: BaseException | ExceptionType,
    exc_types: type[ExceptionT] | tuple[type[ExceptionT], ...],
) -> typing.TypeGuard[ExceptionT]:
    return issubclass(exc, exc_types) if isinstance(exc, type) else isinstance(exc, exc_types)


@dataclasses.dataclass(kw_only=True, frozen=True)
class Error(DataNode, typing.Generic[*ExceptionTs]):
    exception_update: BaseException

    @property
    def exception(self: "Error[*tuple[ExceptionT, ...]]") -> ExceptionT:
        return self.exception_update  # type: ignore

    @classmethod
    def compose(cls, ctx: Context, type_args: TypeArgs) -> typing.Self:
        exception = ctx.exception_update.expect(ComposeError("No exception."))
        exception_types: tuple[ExceptionType, ...] | None = type_args.get(ExceptionTs)

        if exception_types is None or can_catch(exception, exception_types):
            return cls(exception_update=exception)

        raise ComposeError("Foreign exception.")


__all__ = ("Error",)



from telegrinder import node
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.node import NodeRule


class HasText(NodeRule):
    def __init__(self) -> None:
        super().__init__(node.as_node(node.text.Text))


class HasCaption(NodeRule):
    def __init__(self) -> None:
        super().__init__(node.as_node(node.text.Caption))


class Text(ABCRule):
    def __init__(self, texts: str | list[str], /, *, ignore_case: bool = False) -> None:
        if not isinstance(texts, list):
            texts = [texts]
        self.texts = set(texts) if not ignore_case else set(map(str.lower, texts))
        self.ignore_case = ignore_case

    def check(self, text: node.text.Text | node.text.Caption) -> bool:
        return (text if not self.ignore_case else text.lower()) in self.texts


__all__ = ("HasCaption", "HasText", "Text")



import dataclasses
import typing

from fntypes.option import Nothing, Option, Some

from telegrinder.api.api import API
from telegrinder.bot.cute_types import (
    CallbackQueryCute,
    ChatJoinRequestCute,
    ChatMemberUpdatedCute,
    MessageCute,
    PreCheckoutQueryCute,
)
from telegrinder.node.base import ComposeError, DataNode, scalar_node
from telegrinder.node.polymorphic import Polymorphic, impl
from telegrinder.types.objects import Chat, Message, User


@dataclasses.dataclass(kw_only=True, slots=True)
class Source(Polymorphic, DataNode):
    api: API
    from_user: User
    chat: Option[Chat] = dataclasses.field(default_factory=Nothing)
    thread_id: Option[int] = dataclasses.field(default_factory=Nothing)

    @impl
    def compose_message(cls, message: MessageCute) -> typing.Self:
        return cls(
            api=message.api,
            from_user=message.from_user,
            chat=Some(message.chat),
            thread_id=message.message_thread_id,
        )

    @impl
    def compose_callback_query(cls, callback_query: CallbackQueryCute) -> typing.Self:
        return cls(
            api=callback_query.api,
            from_user=callback_query.from_user,
            chat=callback_query.chat,
            thread_id=callback_query.message_thread_id,
        )

    @impl
    def compose_chat_member_updated(cls, chat_member_updated: ChatMemberUpdatedCute) -> typing.Self:
        return cls(
            api=chat_member_updated.api,
            from_user=chat_member_updated.from_user,
            chat=Some(chat_member_updated.chat),
        )

    @impl
    def compose_chat_join_request(cls, chat_join_request: ChatJoinRequestCute) -> typing.Self:
        return cls(
            api=chat_join_request.api,
            from_user=chat_join_request.from_user,
            chat=Some(chat_join_request.chat),
        )

    @impl
    def compose_pre_checkout_query(cls, pre_checkout_query: PreCheckoutQueryCute) -> typing.Self:
        return cls(
            api=pre_checkout_query.api,
            from_user=pre_checkout_query.from_user,
        )

    async def send(self, text: str, **kwargs: typing.Any) -> Message:
        result = await self.api.send_message(
            chat_id=self.chat.map_or(self.from_user.id, lambda chat: chat.id).unwrap(),
            message_thread_id=self.thread_id.unwrap_or_none(),
            text=text,
            **kwargs,
        )
        return result.unwrap()


@scalar_node
class ChatSource:
    @classmethod
    def compose(cls, source: Source) -> Chat:
        return source.chat.expect(ComposeError("Source has no chat."))


@scalar_node
class UserSource:
    @classmethod
    def compose(cls, source: Source) -> User:
        return source.from_user


@scalar_node
class ChatId:
    @classmethod
    def compose(cls, chat: ChatSource) -> int:
        return chat.id


@scalar_node
class UserId:
    @classmethod
    def compose(cls, user: UserSource) -> int:
        return user.id


@scalar_node
class Locale:
    @classmethod
    def compose(cls, user: UserSource) -> str:
        return user.language_code.expect(ComposeError("User has no language code."))


__all__ = (
    "ChatId",
    "ChatSource",
    "Locale",
    "Source",
    "UserId",
    "UserSource",
)



import dataclasses
import os
import pathlib
import re
import typing
from collections import OrderedDict

import libcst as cst

from telegrinder.modules import logger

type APIMethodsMapping = dict[str, cst.FunctionDef]

ANNOTATION_TYPING_ANY: typing.Final[cst.Annotation] = cst.parse_statement("x: typing.Any").body[0].annotation  # type: ignore
DEFAULT_API_METHODS_CLASS_NAME: typing.Final[str] = "APIMethods"
DEFAULT_PATH_CUTE_TYPES: typing.Final[pathlib.Path] = pathlib.Path("telegrinder") / "bot" / "cute_types"


def is_cute_class(node: cst.ClassDef) -> bool:
    if not node.bases or not (node.name.value != "BaseCute" and node.name.value.endswith("Cute")):
        return False

    # Find the base class with the name "BaseCute" or subscriptable (generic) "BaseCute[...]"
    for base in node.bases:
        if isinstance(base.value, cst.Name) and base.value.value == "BaseCute":
            return True

        if (
            isinstance(base.value, cst.Subscript)
            and isinstance(base.value.value, cst.Name)
            and base.value.value == "BaseCute"
        ):
            return True

    return False


def is_decorator_name(decorator_call_node: cst.Call, decorator_name: str) -> bool:
    return isinstance(decorator_call_node.func, cst.Name) and decorator_call_node.func.value == decorator_name


def get_func_params(node: cst.FunctionDef) -> tuple[dict[str, cst.Param], dict[str, cst.Param]]:
    result: tuple[dict[str, cst.Param], dict[str, cst.Param]] = tuple()

    for params in (node.params.params, node.params.kwonly_params):
        params_dct = OrderedDict()
        optional_params = []
        required_params = []

        for param in params:
            if param.name.value in ("cls", "self"):
                params_dct[param.name.value] = param
                continue

            optional_params.append(param) if param.default is not None else required_params.append(param)

        for param in required_params + optional_params:
            params_dct[param.name.value] = param

        result += (params_dct,)  # type: ignore

    return result


def sort_params(params: typing.Iterable[cst.Param]) -> list[cst.Param]:
    by_name = lambda param: param.name.value
    without_defaults = filter(lambda x: x.default is None, params)
    with_defaults = filter(lambda x: x.default is not None, params)
    return sorted(without_defaults, key=by_name) + sorted(with_defaults, key=by_name)


def parse_docstring(docstring: str) -> tuple[str, dict[str, str]]:
    docstring_match = re.match(r"(.*?):param\s+(.*)", docstring, re.DOTALL | re.MULTILINE)
    assert docstring_match
    docstring_match = docstring_match.group(1).strip()
    matches = re.findall(r":param\s+(.*?):(.*)", docstring, re.DOTALL)
    return (docstring_match, {m[0].strip(): m[1].strip() for m in matches})


def prepare_docstring(
    shortcut_func: cst.FunctionDef,
    api_method_func: cst.FunctionDef,
    custom_params: set[str],
) -> str | None:
    if shortcut_func.get_docstring() and (api_m_doc := api_method_func.get_docstring()):
        shortcut_docstring_params = {}
        api_method_docstring, api_method_docstring_params = parse_docstring(api_m_doc)
        shortcut_docstring_s = re.sub(
            r"^Method `.+`",
            f"Shortcut `API.{api_method_func.name.value}()`",
            api_method_docstring,
        )

        for param, docstring in api_method_docstring_params.items():
            if param in custom_params:
                continue
            shortcut_docstring_params[param] = docstring

        for param, docstring in shortcut_docstring_params.items():
            docstring = docstring.replace("         ", "")
            shortcut_docstring_s += f"\n:param {param}: {docstring}"

        return '"""{}"""'.format(shortcut_docstring_s)

    return None


def merge_shortcuts(
    *,
    path_api_methods: str | pathlib.Path,
    path_cute_types: str | pathlib.Path = DEFAULT_PATH_CUTE_TYPES,
    api_methods_class_name: str = DEFAULT_API_METHODS_CLASS_NAME,
) -> None:
    logger.info("Merging shortcut methods with the last changes in Telegram Bot API methods...")

    path_api_methods = pathlib.Path(path_api_methods)
    path_cute_types = pathlib.Path(path_cute_types)

    api_methods_source_tree = cst.parse_module(path_api_methods.read_text(encoding="UTF-8"))
    api_methods_visitor = APIMethodsCollector(api_methods_class_name)
    api_methods_source_tree.visit(api_methods_visitor)

    for path in path_cute_types.rglob("*.py"):
        if path.name == "__init__.py":
            continue

        logger.info("Visit cute module `{}`...", path.stem)

        cute_source_tree = cst.parse_module(path.read_text(encoding="UTF-8"))
        shortcuts_visitor = ShortcutsCollector(api_methods_visitor.api_methods)
        cute_source_tree.visit(shortcuts_visitor)

        transformer = ShortcutsTransformer(
            shortcuts_visitor.shortcuts,
            api_methods_visitor.api_methods,
        )
        modified_cute_tree = cute_source_tree.visit(transformer)
        modified_cute_tree_code = modified_cute_tree.code

        if cute_source_tree.code != modified_cute_tree_code:
            path.write_text(modified_cute_tree_code, encoding="UTF-8")

    logger.info("Formatting cuties with ruff formatter...")
    os.system("ruff format {}".format(path_cute_types))


@dataclasses.dataclass(slots=True, eq=False)
class Shortcut:
    function: cst.FunctionDef
    docstring: str | None = dataclasses.field(default=None, kw_only=True)
    method_name: str = dataclasses.field(kw_only=True)
    custom_params: set[str] = dataclasses.field(default_factory=lambda: set(), kw_only=True)

    def __eq__(self, value: object, /) -> bool:
        return isinstance(value, cst.FunctionDef) and self.function == value


class ShortcutsCollector(cst.CSTVisitor):
    def __init__(self, api_methods: APIMethodsMapping) -> None:
        self.api_methods = api_methods
        self.shortcuts: list[Shortcut] = []

    def visit_ClassDef_body(self, node: cst.ClassDef) -> bool | None:
        """Visit the definition of a class that inherits the `BaseCute` class and the name ends with `Cute`."""
        return is_cute_class(node)

    def visit_FunctionDef_asynchronous(self, node: cst.FunctionDef) -> bool | None:
        """Visit the definition of an async function that are decorated with the `shortcut` decorator."""

        found = False

        for decorator in node.decorators:
            if not isinstance(decorator.decorator, cst.Call):
                continue

            if is_decorator_name(decorator.decorator, "staticmethod"):
                return False

            if is_decorator_name(decorator.decorator, "shortcut"):
                kwargs = {}

                for arg in decorator.decorator.args:
                    if isinstance(arg.value, cst.SimpleString):
                        kwargs["method_name"] = arg.value.value.removeprefix('"').removesuffix('"')
                    elif isinstance(arg.value, cst.Set):
                        kwargs["custom_params"] = {
                            e.value.value.removeprefix('"').removesuffix('"')
                            for e in arg.value.elements
                            if isinstance(e.value, cst.SimpleString)
                        }

                found = True
                self.shortcuts.append(
                    Shortcut(
                        node,
                        docstring=prepare_docstring(
                            node,
                            self.api_methods[kwargs["method_name"]],
                            kwargs.get("custom_params", set()),
                        ),
                        **kwargs,
                    ),
                )

        return found


class APIMethodsCollector(cst.CSTVisitor):
    def __init__(self, api_class_methods_name: str) -> None:
        self.api_class_methods_name = api_class_methods_name
        self.api_methods: APIMethodsMapping = {}

    def visit_ClassDef_body(self, node: cst.ClassDef) -> bool | None:
        """Visit the definition of a class that the name == `self.api_class_methods_name` and find definitions of an async function."""
        if node.name.value != self.api_class_methods_name:
            return False

        for statement in node.body.body:
            if isinstance(statement, cst.FunctionDef) and statement.asynchronous is not None:
                self.api_methods[statement.name.value] = statement

        return True


class ShortcutsTransformer(cst.CSTTransformer):
    def __init__(self, shortcuts: list[Shortcut], api_methods: APIMethodsMapping) -> None:
        self.shortcuts = shortcuts
        self.api_methods = api_methods

    def visit_ClassDef_body(self, node: cst.ClassDef) -> bool | None:
        return is_cute_class(node)

    def visit_FunctionDef_asynchronous(self, node: cst.FunctionDef) -> bool | None:
        return node in self.shortcuts

    def leave_FunctionDef(
        self,
        original_node: cst.FunctionDef,
        updated_node: cst.FunctionDef,
    ) -> cst.FunctionDef:
        if typing.cast("Shortcut", original_node) in self.shortcuts:
            shortcut = self.shortcuts.pop(self.shortcuts.index(typing.cast("Shortcut", original_node)))
            shortcut_args, shortcut_kwargs = get_func_params(shortcut.function)
            api_method_args, api_method_kwargs = get_func_params(self.api_methods[shortcut.method_name])

            for apiargs, shortcutargs in (
                (api_method_args, shortcut_args),
                (api_method_kwargs, shortcut_kwargs),
            ):
                for name, param in apiargs.items():
                    if (
                        name in ("cls", "self")
                        or name in shortcut.custom_params
                        or param.annotation is None
                        or name in shortcut_args
                        or (param.default is not None and not isinstance(param.default, cst.Name))
                    ):
                        continue

                    shortcutargs[name] = param

            params = [
                shortcut_args.pop("cls", None) or shortcut_args.pop("self", None) or cst.Param(cst.Name("self")),
            ]
            return updated_node.with_changes(
                params=cst.Parameters(
                    params=params + sort_params(shortcut_args.values()),
                    kwonly_params=sort_params(shortcut_kwargs.values()),
                    star_arg=cst.ParamStar() if shortcut_kwargs else cst.MaybeSentinel.DEFAULT,
                    star_kwarg=cst.Param(
                        cst.Name("other"),
                        annotation=ANNOTATION_TYPING_ANY,
                        comma=cst.Comma(cst.TrailingWhitespace()),  # type: ignore
                    ),
                ),
            )

        return updated_node

    def leave_SimpleString(
        self,
        original_node: cst.SimpleString,
        updated_node: cst.SimpleString,
    ) -> cst.SimpleString:
        for shortcut in self.shortcuts:
            if (
                shortcut.docstring
                and shortcut.function.get_docstring(clean=False) == original_node.evaluated_value
            ):
                return updated_node.with_changes(value=shortcut.docstring)

        return updated_node


__all__ = ("merge_shortcuts",)



from __future__ import annotations

import asyncio
import datetime
import typing

from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.view.base import BaseView
from telegrinder.bot.dispatch.waiter_machine.middleware import WaiterMiddleware
from telegrinder.bot.dispatch.waiter_machine.short_state import (
    ShortState,
    ShortStateContext,
)
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.tools.global_context.builtin_context import TelegrinderContext
from telegrinder.tools.lifespan import Lifespan
from telegrinder.tools.limited_dict import LimitedDict

from .actions import WaiterActions
from .hasher import Hasher

type Storage[Event: BaseCute, HasherData] = dict[
    Hasher[Event, HasherData],
    LimitedDict[typing.Hashable, ShortState[Event]],
]
type HasherWithData[Event: BaseCute, Data] = tuple[Hasher[Event, Data], Data]

_NODEFAULT: typing.Any = object()
MAX_STORAGE_SIZE: typing.Final[int] = 10000
ONE_MINUTE: typing.Final[datetime.timedelta] = datetime.timedelta(minutes=1)
WEEK: typing.Final[datetime.timedelta] = datetime.timedelta(days=7)
CONTEXT: typing.Final[TelegrinderContext] = TelegrinderContext()


async def clear_wm_storage_worker(wm: WaiterMachine, interval: float) -> None:
    await wm.clear_storage()
    await asyncio.sleep(interval)


def unpack_to_context(context: Context, /) -> tuple[Context]:
    return (context,)


def no_unpack(_: Context, /) -> tuple[()]:
    return ()


class ContextUnpackProto[*Ts](typing.Protocol):
    __name__: str

    def __call__(self, context: Context, /) -> tuple[*Ts]: ...


class WaiterMachine:
    def __init__(
        self,
        dispatch: ABCDispatch | None = None,
        *,
        max_storage_size: int = MAX_STORAGE_SIZE,
        base_state_lifetime: datetime.timedelta = WEEK,
        clear_storage_interval: datetime.timedelta = ONE_MINUTE,
    ) -> None:
        self.dispatch = dispatch
        self.max_storage_size = max_storage_size
        self.base_state_lifetime = base_state_lifetime
        self.storage: Storage[typing.Any, typing.Any] = {}

        CONTEXT.loop_wrapper.add_task(
            clear_wm_storage_worker(self, clear_storage_interval.total_seconds()),
        )

    def __repr__(self) -> str:
        return "<{}: with {} storage items and max_storage_size={}, base_state_lifetime={!r}>".format(
            self.__class__.__name__,
            len(self.storage),
            self.max_storage_size,
            self.base_state_lifetime,
        )

    async def drop_all(self) -> None:
        """Drops all waiters in storage."""
        for hasher in self.storage.copy():
            for ident, short_state in self.storage[hasher].items():
                if short_state.context:
                    await self.drop(hasher, data=ident)
                else:
                    await short_state.cancel()

            del self.storage[hasher]

    async def drop[Event: BaseCute, HasherData](
        self,
        hasher: Hasher[Event, HasherData],
        data: HasherData,
        **context: typing.Any,
    ) -> None:
        if hasher not in self.storage:
            raise LookupError("No record of hasher {!r} found.".format(hasher))

        waiter_id: typing.Hashable = hasher.get_hash_from_data(data).expect(
            RuntimeError("Couldn't create hash from data"),
        )
        short_state = self.storage[hasher].pop(waiter_id, None)
        if short_state is None:
            raise LookupError(
                "Waiter with identificator {} is not found for hasher {!r}.".format(waiter_id, hasher)
            )

        if on_drop := short_state.actions.get("on_drop"):
            on_drop(short_state, **context)

        await short_state.cancel()

    async def wait[Event: BaseCute, HasherData](
        self,
        hasher: Hasher[Event, HasherData] | HasherWithData[Event, HasherData],
        data: HasherData = _NODEFAULT,
        *,
        filter: ABCRule | None = None,
        release: ABCRule | None = None,
        lifetime: datetime.timedelta | float | None = None,
        lifespan: Lifespan | None = None,
        **actions: typing.Unpack[WaiterActions[Event]],
    ) -> ShortStateContext[Event]:
        if isinstance(lifetime, int | float):
            lifetime = datetime.timedelta(seconds=lifetime)

        lifespan = lifespan or Lifespan()
        event = asyncio.Event()
        short_state = ShortState(
            event,
            actions,
            release=release,
            filter=filter,
            lifetime=lifetime or self.base_state_lifetime,
        )

        hasher, data = hasher if not isinstance(hasher, Hasher) else (hasher, data)
        assert data is not _NODEFAULT, "Hasher requires data."
        waiter_hash = hasher.get_hash_from_data(data).expect(RuntimeError("Hasher couldn't create hash."))

        if hasher not in self.storage:
            if self.dispatch:
                view: BaseView = self.dispatch.get_view(hasher.view_class).expect(
                    RuntimeError(f"View {hasher.view_class.__name__!r} is not defined in dispatch."),
                )
                view.middlewares.insert(0, WaiterMiddleware(self, hasher))

            self.storage[hasher] = LimitedDict(maxlimit=self.max_storage_size)

        if (deleted_short_state := self.storage[hasher].set(waiter_hash, short_state)) is not None:
            await deleted_short_state.cancel()

        async with lifespan:
            await event.wait()

        self.storage[hasher].pop(waiter_hash, None)

        if short_state.context is None:
            raise LookupError("No context in short_state.")
        return short_state.context

    async def wait_many[Event: BaseCute[typing.Any], Data, *Ts](
        self,
        *hashers: HasherWithData[Event, Data],
        filter: ABCRule | None = None,
        release: ABCRule | None = None,
        lifetime: datetime.timedelta | float | None = None,
        lifespan: Lifespan | None = None,
        unpack: ContextUnpackProto[*Ts] = unpack_to_context,
        **actions: typing.Unpack[WaiterActions[Event]],
    ) -> tuple[HasherWithData[Event, Data], Event, *Ts]:
        if isinstance(lifetime, int | float):
            lifetime = datetime.timedelta(seconds=lifetime)

        lifespan = lifespan or Lifespan()
        event = asyncio.Event()
        short_state = ShortState(
            event,
            actions,
            release=release,
            filter=filter,
            lifetime=lifetime or self.base_state_lifetime,
        )
        waiter_hashes: dict[Hasher[Event, Data], typing.Hashable] = {}

        for hasher, data in hashers:
            waiter_hash = hasher.get_hash_from_data(data).expect(RuntimeError("Hasher couldn't create hash."))

            if hasher not in self.storage:
                if self.dispatch:
                    view = self.dispatch.get_view(hasher.view_class).expect(
                        RuntimeError(f"View {hasher.view_class.__name__!r} is not defined in dispatch."),
                    )
                    view.middlewares.insert(0, WaiterMiddleware(self, hasher))
                self.storage[hasher] = LimitedDict(maxlimit=self.max_storage_size)

            if (deleted_short_state := self.storage[hasher].set(waiter_hash, short_state)) is not None:
                await deleted_short_state.cancel()

            waiter_hashes[hasher] = waiter_hash

        async with lifespan:
            await event.wait()

        if short_state.context is None:
            raise LookupError("No context in short_state.")

        initiator = short_state.context.context.get("initiator")
        if initiator is None:
            raise LookupError("Initiator not found in short_state context.")

        for hasher, waiter_hash in waiter_hashes.items():
            self.storage[hasher].pop(waiter_hash, None)

        return (
            initiator,
            short_state.context.event,  # type: ignore
            *unpack(short_state.context.context),
        )

    async def clear_storage(self) -> None:
        """Clears storage."""
        now = datetime.datetime.now()

        for hasher in self.storage:
            for ident, short_state in self.storage.get(hasher, {}).copy().items():
                if short_state.expiration_date is not None and now > short_state.expiration_date:
                    await self.drop(hasher, data=ident, force=True)


__all__ = ("WaiterMachine",)



from telegrinder.tools.singleton.abc import ABCSingleton, ABCSingletonMeta
from telegrinder.tools.singleton.singleton import Singleton, SingletonMeta

__all__ = ("ABCSingleton", "ABCSingletonMeta", "Singleton", "SingletonMeta")



from __future__ import annotations

import dataclasses
import typing
from copy import deepcopy

from telegrinder.tools.keyboard.base import BaseKeyboard, DictStrAny
from telegrinder.tools.keyboard.data import KeyboardModel
from telegrinder.types.objects import (
    InlineKeyboardMarkup,
    ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
)

if typing.TYPE_CHECKING:
    from telegrinder.tools.keyboard.abc import RawKeyboard


@dataclasses.dataclass(kw_only=True, frozen=True)
class KeyboardModelMixin(KeyboardModel if typing.TYPE_CHECKING else object):
    keyboard_model: KeyboardModel = dataclasses.field(init=False)

    if not typing.TYPE_CHECKING:
        keyboard_model = dataclasses.field(
            default_factory=lambda: KeyboardModel(keyboard=[[]]),
            repr=False,
        )

    @property
    def keyboard(self) -> RawKeyboard:
        return self.keyboard_model.keyboard  # type: ignore

    @property
    def is_persistent(self) -> bool:
        return self.keyboard_model.is_persistent

    @property
    def one_time_keyboard(self) -> bool:
        return self.keyboard_model.one_time_keyboard

    @property
    def resize_keyboard(self) -> bool:
        return self.keyboard_model.resize_keyboard

    @property
    def is_selective(self) -> bool:
        return self.keyboard_model.selective

    @property
    def input_field_placeholder(self) -> str | None:
        return self.keyboard_model.input_field_placeholder


@dataclasses.dataclass(frozen=True, kw_only=True)
class Keyboard(KeyboardModelMixin, BaseKeyboard):
    if not typing.TYPE_CHECKING:

        def __init__(self, **kwargs):
            super().__init__(keyboard_model=kwargs.pop("keyboard_model", None) or KeyboardModel(**kwargs))

    def dict(self) -> DictStrAny:
        return dataclasses.asdict(self.keyboard_model) | {"keyboard": [row for row in self.keyboard if row]}

    def get_markup(self) -> ReplyKeyboardMarkup:
        return ReplyKeyboardMarkup.from_dict(self.dict())

    def copy(self, **with_changes: typing.Any) -> typing.Self:
        return dataclasses.replace(
            self,
            keyboard_model=dataclasses.replace(
                self.keyboard_model,
                keyboard=deepcopy(self.keyboard_model.keyboard),
                **with_changes,
            ),
        )

    def get_keyboard_remove(self) -> ReplyKeyboardRemove:
        return ReplyKeyboardRemove.from_data(
            remove_keyboard=True,
            selective=self.keyboard_model.selective,
        )

    def resize(self) -> typing.Self:
        return self.copy(resize_keyboard=True)

    def one_time(self) -> typing.Self:
        return self.copy(one_time_keyboard=True)

    def selective(self) -> typing.Self:
        return self.copy(selective=True)

    def persistent(self) -> typing.Self:
        return self.copy(is_persistent=True)

    def no_resize(self) -> typing.Self:
        return self.copy(resize_keyboard=False)

    def no_one_time(self) -> typing.Self:
        return self.copy(one_time_keyboard=False)

    def no_selective(self) -> typing.Self:
        return self.copy(selective=False)

    def no_persistent(self) -> typing.Self:
        return self.copy(is_persistent=False)

    def placeholder(self, value: str | None, /) -> typing.Self:
        return self.copy(input_field_placeholder=value)


@dataclasses.dataclass(frozen=True)
class InlineKeyboard(BaseKeyboard):
    keyboard: RawKeyboard = dataclasses.field(init=False)

    if not typing.TYPE_CHECKING:
        keyboard = dataclasses.field(
            default_factory=lambda: [[]],
            repr=False,
        )

    def dict(self) -> DictStrAny:
        return dict(inline_keyboard=[row for row in self.keyboard if row])

    def get_markup(self) -> InlineKeyboardMarkup:
        return InlineKeyboardMarkup.from_dict(self.dict())

    def copy(self, **_: typing.Any) -> typing.Self:
        return dataclasses.replace(self, keyboard=deepcopy(self.keyboard))


__all__ = ("InlineKeyboard", "Keyboard")
